"""
Configures a logger to be used across all of pybedtools
"""
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
ch = logging.StreamHandler()
ch.setLevel(logging.WARNING)
formatter = logging.Formatter("%(name)s [%(levelname)s]: %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)



#!/usr/bin/env python
"""
Make a pie chart where peaks fall in annotations; see \
:mod:`pybedtools.contrib.Classifier` for more flexibility.

The results here are similar to CEAS (http://liulab.dfci.harvard.edu/CEAS/).

However, multi-featuretype classes are reported.  That is, if a peak falls in
an exon in one isoform and an intron in another isoform, the class is "exon,
intron".
"""

import sys
from six.moves import urllib
import argparse
import pybedtools
from collections import defaultdict


def make_pie(
    bed, gff, stranded=False, out="out.png", include=None, exclude=None, thresh=0
):

    a = pybedtools.BedTool(bed)
    b = pybedtools.BedTool(gff).remove_invalid()

    c = a.intersect(b, wao=True, s=stranded, stream=True)

    # So we can grab just `a` features later...
    afields = a.field_count()

    # Where we can find the featuretype in the -wao output.  Assumes GFF.
    type_idx = afields + 2

    # 3 different code paths depending on include/exclude to cut down on
    # if/else checks.
    #
    # For un-included featuretypes, put them in the '.' category (unnannotated)
    if include and exclude:
        raise ValueError("Can only specify one of `include` or `exclude`.")
    d = defaultdict(set)
    if include:
        for feature in c:
            featuretype = feature[type_idx]
            key = "\t".join(feature[:afields])
            if featuretype in include:
                d[key].update([featuretype])
            else:
                d[key].update(["."])
    elif exclude:
        for feature in c:
            featuretype = feature[type_idx]
            key = "\t".join(feature[:afields])
            if featuretype not in exclude:
                d[key].update([featuretype])
            else:
                d[key].update(["."])
    else:
        for feature in c:
            featuretype = feature[type_idx]
            key = "\t".join(feature[:afields])
            d[key].update([featuretype])

    def labelmaker(x):
        x.difference_update(".")
        label = []
        for i in list(x):
            if i == "three_prime_UTR":
                i = "3' UTR"
            if i == "five_prime_UTR":
                i = "5' UTR"
            label.append(i)
        return ", ".join(sorted(label))

    # Prepare results for Google Charts API
    npeaks = float(len(d))
    count_d = defaultdict(int)
    for peak, featuretypes in list(d.items()):
        if featuretypes == set("."):
            featuretype = "unannotated"
        else:
            featuretype = labelmaker(featuretypes)
        count_d[featuretype] += 1

    results = list(count_d.items())
    results.sort(key=lambda x: x[1])
    labels, counts = list(zip(*results))

    labels = []
    counts_to_use = []
    for label, count in results:
        perc = count / npeaks * 100
        if perc > thresh:
            labels.append("%s: %s (%.1f%%)" % (label, count, perc))
            counts_to_use.append(perc)

    # Set up the Gchart data
    data = {
        "cht": "p",
        "chs": "750x350",
        "chd": "t:" + ",".join(map(str, counts_to_use)),
        "chl": "|".join(labels),
    }

    # Encode it correctly
    encoded_data = urllib.parse.urlencode(data)

    # Send request and get data; write to file
    url = "https://chart.googleapis.com/chart?"
    req = urllib.request.Request(url, encoded_data)
    response = urllib.request.urlopen(req)
    f = open(out, "w")
    f.write(response.read())
    f.close()


def main():
    """
    Make a pie chart of features overlapping annotations (e.g., peaks in
    introns, exons, etc)
    """
    ap = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    ap.add_argument("--bed", help="BED file of e.g. peaks")
    ap.add_argument("--gff", help="GFF file of e.g. annotations")
    ap.add_argument("--out", default="out.png", help="Output PNG file")
    ap.add_argument(
        "--stranded", action="store_true", help="Use strand-specific intersections"
    )
    ap.add_argument("--include", nargs="*", help="Featuretypes to include")
    ap.add_argument("--exclude", nargs="*", help="Featuretypes to exclude")
    ap.add_argument(
        "--thresh",
        type=float,
        help="Threshold percentage below which output will be " "suppressed",
    )
    ap.add_argument(
        "--test",
        action="store_true",
        help="Run test, overwriting all other args. Result will "
        'be "out.png" in current directory.',
    )
    args = ap.parse_args()

    if not (args.bed and args.gff) and not args.test:
        ap.print_help()
        sys.exit(1)

    if not args.test:
        if args.include and args.exclude:
            raise ValueError("Cannot specify both --include and --exclude")

        make_pie(
            bed=args.bed,
            gff=args.gff,
            out=args.out,
            thresh=args.thresh,
            stranded=args.stranded,
            include=args.include,
            exclude=args.exclude,
        )
    else:
        make_pie(
            bed=pybedtools.example_filename("gdc.bed"),
            gff=pybedtools.example_filename("gdc.gff"),
            stranded=True,
            out="out.png",
            include=["exon", "CDS", "intron", "five_prime_UTR", "three_prime_UTR"],
        )


if __name__ == "__main__":
    import doctest

    if doctest.testmod(optionflags=doctest.ELLIPSIS).failed == 0:
        main()



"""
Interface between pybedtools and the R package VennDiagram.

Rather than depend on the user to have rpy2 installed, this simply writes an
R script that can be edited and tweaked by the user before being run in R.
"""
import os
import string
import pybedtools
import six
from pybedtools import helpers
import subprocess
from collections import OrderedDict

# really just fill in x and filename...leave the rest up to the user.
#
# Note that the closing parentheses is missing -- that's so the user can add
# kwargs from the calling function
template = string.Template(
    """
library(VennDiagram)
venn.diagram(
    x=$x,
    filename=$filename,
    category.names = $names
"""
)


def _list_to_R_syntax(x):
    """
    Convert items in `x` to a string, and replace tabs with pipes in Interval
    string representations.  Put everything into an R vector and return as one
    big string.
    """
    items = []
    for i in x:
        if isinstance(i, pybedtools.Interval):
            i = str(i).replace("\t", "|")
        items.append('"%s"' % i)
    return "c(%s)" % ",".join(items)


def _dict_to_R_named_list(d):
    """
    Calls _list_to_R_syntax for each item.  Returns one big string.
    """
    items = []
    for key, val in list(d.items()):
        items.append('"%s" = %s' % (key, _list_to_R_syntax(val)))
    return "list(%s)" % ", ".join(items)


def truncator(feature):
    """
    Convert a feature of any format into a BED3 format.
    """
    return pybedtools.create_interval_from_list(
        [feature.chrom, str(feature.start), str(feature.stop)]
    )


def cleaned_intersect(items):
    """
    Perform interval intersections such that the end products have identical \
    features for overlapping intervals.

    The VennDiagram package does *set* intersection, not *interval*
    intersection.  So the goal here is to represent intersecting intervals as
    intersecting sets of strings.

    Doing a simple BEDTools intersectBed call doesn't do the trick (even with
    the -u argument).  As a concrete example, what would the string be for an
    intersection of the feature "chr1:1-100" in file `x` and "chr1:50-200" in
    file `y`?

    The method used here is to substitute the intervals in `y` that overlap `x`
    with the corresponding elements in `x`.  This means that in the resulting
    sets, the overlapping features are identical.  To follow up with the
    example, both `x` and `y` would have an item "chr1:50-200" in their sets,
    simply indicating *that* one interval overlapped.

    Venn diagrams are not well suited for nested overlaps or multi-overlaps.
    To illustrate, try drawing the 2-way Venn diagram of the following two
    files. Specifically, what number goes in the middle -- the number of
    features in `x` that intersect `y` (1) or the number of features in `y`
    that intersect `x` (2)?::

        x:
            chr1  1  100
            chr1 500 6000

        y:
            chr1 50 100
            chr1 80 200
            chr9 777 888

    In this case, this function will return the following sets::

        x:
            chr1:1-100
            chr1:500-6000

        y:
            chr1:1-100
            chr9:777-888

    This means that while `x` does not change in length, `y` can.  For example,
    if there are 2 features in `x` that overlap one feature in `y`, then `y`
    will gain those two features in place of its single original feature.

    This strategy is extended for multiple intersections -- see the source for
    details.
    """
    if len(items) == 2:
        x = items[0].each(truncator).saveas()
        y = items[1].each(truncator).saveas()

        # Combine the unique-to-y intervals with the shared-with-x intervals.
        # Since x is first in x+y, resulting features are from x.
        new_y = (y - x).cat(x + y)
        return x, new_y

    if len(items) == 3:
        x = items[0].each(truncator).saveas()
        y = items[1].each(truncator).saveas()
        z = items[2].each(truncator).saveas()

        # Same as above.  Don't care about z yet; this means that y will not
        # change because of z.
        new_y = (y - x).cat(x + y)

        # Combine:
        #  unique-to-z
        #  shared-with-any-x
        #  shared-with-unique-to-y
        new_z = (z - y - x).cat(x + z).cat((y - x) + z)
        return x, new_y, new_z

    if len(items) == 4:
        x = items[0].each(truncator).saveas()
        y = items[1].each(truncator).saveas()
        z = items[2].each(truncator).saveas()
        q = items[3].each(truncator).saveas()

        # Same as 2-way
        new_y = (y - x).cat(x + y)

        # Same as 3-way
        new_z = (z - y - x).cat(x + z).cat((y - x) + z)

        # Combine:
        #  unique-to-q
        #  shared-with-any-x
        #  shared-with-unique-to-y
        #  shared-with-unique-to-z
        new_q = (q - z - y - x).cat(x + q).cat((y - x) + q).cat((z - y - x) + q)

        return x, new_y, new_z, new_q


def venn_maker(
    beds,
    names=None,
    figure_filename=None,
    script_filename=None,
    additional_args=None,
    run=False,
):
    """
    Given a list of interval files, write an R script to create a Venn \
    diagram of overlaps (and optionally run it).

    The R script calls the venn.diagram function of the R package VennDiagram
    for extremely flexible Venn and Euler diagram creation.  Uses
    `cleaned_intersect()` to create string representations of shared intervals.

    `beds` is a list of up to 4 filenames or BedTools.

    `names` is a list of names to use for the Venn diagram, in the same order
    as `beds`. Default is "abcd"[:len(beds)].

    `figure_filename` is the TIFF file to save the figure as.

    `script_filename` is the optional filename to write the R script to

    `additional_args` is list that will be inserted into the R script,
    verbatim.  For example, to use scaled Euler diagrams with different colors,
    use::

        additional_args = ['euler.d=TRUE',
                           'scaled=TRUE',
                           'cat.col=c("red","blue")']

    If `run` is True, then assume R is installed, is on the path, and has
    VennDiagram installed . . . and run the script.  The resulting filename
    will be saved as `figure_filename`.
    """

    if figure_filename is None:
        figure_filename = "NULL"
    else:
        figure_filename = '"%s"' % figure_filename

    if names is None:
        names = "abcd"[: len(beds)]

    _beds = []
    for bed in beds:
        if not isinstance(bed, pybedtools.BedTool):
            bed = pybedtools.BedTool(bed)
        _beds.append(bed)

    cleaned = cleaned_intersect(_beds)
    results = OrderedDict(list(zip(names, cleaned)))

    s = template.substitute(
        x=_dict_to_R_named_list(results),
        filename=figure_filename,
        names=_list_to_R_syntax(names),
    )
    if additional_args:
        s += "," + ", ".join(additional_args)

    s += ")"

    if not script_filename:
        fn = pybedtools.BedTool._tmp()
    else:
        fn = script_filename

    fout = open(fn, "w")
    fout.write(s)
    fout.close()

    out = fn + ".Rout"
    if run:

        if not pybedtools.settings._R_installed:
            helpers._check_for_R()

        cmds = [os.path.join(pybedtools.settings._R_path, "R"), "CMD", "BATCH", fn, out]
        p = subprocess.Popen(cmds, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        stdout, stderr = p.communicate()
        if stdout or stderr:
            print("stdout:", stdout)
            print("stderr:", stderr)

    if not script_filename:
        return s

    return None



"""
tests for contrib module
"""
import sys
import os
import pybedtools
from pybedtools import Interval

# from pybedtools.contrib import Classifier
from .tfuncs import setup_module, teardown_module, testdir, test_tempdir, unwriteable

# model for gdc.
# chr2L, starts at 1 and spacing is 10bp.
# import gdc
# g = gdc.GenomeModel(chrom_start=1,
#                     chrom='chr2L',
#                     scalar=10,
#                     read_length=5)
"""
#         10       20        30
#123456789012345678901234567890
   >===||||||~~~~|||==@    #mRNA_xs2_g2_+
               <|||||||@   #tRNA_t2_t2_-
$+     -      --     +     #
$      +      + +          #
"""
# ^     ^      ^^^    ^
# |     |      |      |- STRANDED: exon, UTR, mRNA, gene
#      |      |         UNSTRANDED: exon, UTR, mRNA, tRNA, gene, CDS (cause gdc considers tRNA's exon as CDS)
# |     |      |||- STRANDED: intron, mRNA, gene
# |     |      |    UNSTRANDED: intron, mRNA, gene, tRNA, CDS (because GDC considers the tRNA's exon as CDS)
# |     |      ||- STRANDED: none
# |     |      |   UNSTRANDED: intron, mRNA, gene
# |     |      |-STRANDED: +: intron, mRNA, gene; -: none
# |     |        UNSTRANDED: intron, mRNA, gene
# |     |- STRANDED: +: exon, CDS, mRNA, gene; -: none
# |        UNSTRANDED: exon, CDS, mRNA, gene
# - STRANDED: none
#  UNSTRANDED: none


def fix(x):
    """
    Replaces spaces with tabs, removes spurious newlines, and lstrip()s each
    line. Makes it really easy to create BED files on the fly for testing and
    checking.
    """
    s = ""
    for i in x.splitlines():
        i = i.lstrip()
        if i.endswith("\t"):
            add_tab = "\t"
        else:
            add_tab = ""
        if len(i) == 0:
            continue
        i = i.split()
        i = "\t".join(i) + add_tab + "\n"
        s += i
    return s


def _classifier():

    c = Classifier(
        bed=pybedtools.example_filename("gdc.bed"),
        annotations=pybedtools.example_filename("gdc.gff"),
    )
    c.classify()

    bed = pybedtools.example_bedtool("gdc.bed")

    assert c.class_counts == {
        frozenset(["UTR", "exon", "mRNA", "CDS", "tRNA", "gene"]): 1,
        frozenset(["intron", "gene", "mRNA"]): 3,
        frozenset([]): 1,
        frozenset(["gene", "exon", "mRNA", "CDS"]): 2,
        frozenset(["exon", "mRNA", "CDS", "tRNA", "intron", "gene"]): 1,
    }

    assert c.feature_classes == {
        bed[0]: set(["."]),
        bed[1]: set(["gene", "exon", "mRNA", "CDS"]),
        bed[2]: set(["intron", "gene", "mRNA"]),
        bed[3]: set(["intron", "gene", "mRNA"]),
        bed[4]: set(["tRNA", "UTR", "exon", "mRNA", "CDS", "gene"]),
        bed[5]: set(["gene", "exon", "mRNA", "CDS"]),
        bed[6]: set(["intron", "gene", "mRNA"]),
        bed[7]: set(["tRNA", "intron", "exon", "mRNA", "CDS", "gene"]),
    }

    print("use these indexes for debugging")
    for i, f in enumerate(bed):
        print(i, f)

    for k, v in list(c.class_features.items()):
        print(k)
        for i in v:
            print("\t" + str(i))

    assert c.class_features == {
        frozenset([]): [bed[0]],
        frozenset(["intron", "gene", "mRNA"]): [bed[6], bed[2], bed[3]],
        frozenset(["gene", "exon", "mRNA", "CDS"]): [bed[5], bed[1]],
        frozenset(["UTR", "exon", "mRNA", "CDS", "tRNA", "gene"]): [bed[4]],
        frozenset(["exon", "mRNA", "CDS", "tRNA", "intron", "gene"]): [bed[7]],
    }


def test_cleaned_intersect():
    x = pybedtools.BedTool(
        """
    chr1 1 10      1
    chr1 20 30     2
    chr1 100 120   3
    """,
        from_string=True,
    )
    y = pybedtools.BedTool(
        """
    chr1 2 7       4
    chr1 110 120   5
    chr1 200 210   6
    """,
        from_string=True,
    )
    z = pybedtools.BedTool(
        """
    chr1 25 40     7
    chr1 190 205   8
    chr1 1000 1001 9
    """,
        from_string=True,
    )

    # Two-way test
    #
    x2, y2 = pybedtools.contrib.venn_maker.cleaned_intersect([x, y])

    # x should be the same -- 1, 2, 3
    # y should have 1, 3, 6

    assert x2 == fix(
        """
    chr1 1 10
    chr1 20 30
    chr1 100 120
    """
    )

    assert y2 == fix(
        """
    chr1 1 10
    chr1 100 120
    chr1 200 210"""
    )

    # Three-way test
    #
    x3, y3, z3 = pybedtools.contrib.venn_maker.cleaned_intersect([x, y, z])

    # x should be the same -- 1, 2, 3
    # y should have 1, 3, 6
    # z should have 2, 6

    assert x3 == fix(
        """
    chr1 1 10
    chr1 20 30
    chr1 100 120
    """
    )

    assert y3 == fix(
        """
    chr1 1 10
    chr1 100 120
    chr1 200 210"""
    )

    assert z3 == fix(
        """
    chr1 20 30
    chr1 200 210
    chr1 1000 1001"""
    )

    try:
        pybedtools.helpers._check_for_R()
        print(
            pybedtools.contrib.venn_maker.venn_maker(
                beds=[x, y, z],
                names=["x", "y", "z"],
                figure_filename="out.tiff",
                additional_args=[
                    "euler.d=TRUE",
                    "scaled=TRUE",
                    'fill=c("red","blue", "orange")',
                ],
                run=True,
            )
        )
    except ValueError:
        sys.stderr.write("R installation not found; skipping test")

    if os.path.exists("out.tiff"):
        os.unlink("out.tiff")


def test_venn_maker_3way_1empty():
    # Fix issue #95. The problem was that BedTool.cat() failed when checking
    # field counts on an empty file.  The fix was to make cat() aware of empty
    # files when checking field num and field type.
    a = pybedtools.BedTool(
        """
    chr1 10 100
    chr2 10 100""",
        from_string=True,
    )
    b = pybedtools.BedTool(
        """
    chr1 12 80""",
        from_string=True,
    )
    c = pybedtools.BedTool(
        """
    chr2 20 40""",
        from_string=True,
    )
    try:
        pybedtools.contrib.venn_maker.venn_maker(
            [a, b, c], run=True, figure_filename="t.tiff"
        )
    except ValueError:
        print("R not installed, skipping test")
        # os.unlink('t.tiff')



# distutils: language = c++
from cbedtools cimport Interval
from cbedtools import create_interval_from_list


cpdef extend_fields(Interval feature, int n):
    """
    Pads the fields of the feature with "." to a total length of `n` fields,
    """
    fields = feature.fields[:]
    while len(fields) < n:
        fields.append('.')
    i = create_interval_from_list(fields)

    if n > 4 and (i[4] == '.'):
        i[4] = '0'
    if n > 6 and (i[6] == '.'):
        i[6] = str(i.start)
    if n > 7 and (i[7] == '.'):
        i[7] = str(i.stop)
    if n > 8 and (i[8] == '.'):
        i[8] = '0,0,0'
    return i



cpdef center(Interval feature, int width=100):
    """
    Return the *width* bp from the center of a feature.  If a feature is
    smaller than *width*, then return the entire feature.
    """
    if len(feature) < width:
        return feature
    cdef int start = feature.start
    cdef int stop = feature.stop
    cdef int center = start + (stop - start) / 2
    halfwidth = width / 2
    feature.start = center - halfwidth
    if feature.start < 1:
        feature.start = 1
    if halfwidth == 0:
        halfwidth = 1
    feature.stop = center + halfwidth
    return feature


cpdef midpoint(Interval feature):
    """
    Specialized version of `center()` that just returns the single-bp midpoint
    """
    start = feature.start + (feature.stop - feature.start) / 2
    stop = start + 1
    feature.start = start
    feature.stop = stop
    return feature


cpdef greater_than(Interval feature, int size=100):
    """
    Return True if feature length > *size*
    """
    return len(feature) > size


cpdef less_than(Interval feature, int size=100):
    """
    Return True if feature length < *size*
    """
    return len(feature) < size


cpdef normalized_to_length(Interval feature, int idx=4, float scalar=0.001):
    """
    Normalizes the value at feature[idx] to the feature's length, in kb.

    *idx*, by default, is the score field for a BED file, but specify any
    integer.

    The value at *idx* will be replaced with its scaled value.

    *scalar* will be multiplied by the value at *idx*, by default this is
    0.001, or per kb.

    Useful for calculating RPKM after running intersect with counts
    """
    feature[idx] = str(float(feature[idx]) * scalar / len(feature))
    return feature


cpdef rename(Interval feature, str name):
    """
    Forces a rename of all features, e.g., for renaming everything in a file
    'exon'
    """
    feature.name = name
    return feature


cpdef bedgraph_scale(Interval feature, float scalar):
    feature[3] = str(float(feature[3]) * scalar)
    return feature


cpdef TSS(Interval feature, int upstream=500, int downstream=500, add_to_name=None, genome=None):
    """
    Alias for five_prime.
    """
    return star_prime(feature, upstream, downstream, prime=5,
                      add_to_name=add_to_name, genome=genome)


cdef star_prime(Interval feature, int upstream=500, int downstream=500, int prime=5,
                 add_to_name=None, genome=None):

    if prime == 5:
        if feature.strand == '-':
            start = feature.stop - downstream
            stop = feature.stop + upstream
        else:
            start = feature.start - upstream
            stop = feature.start + downstream
    elif prime == 3:
        if feature.strand == '-':
            start = feature.start - downstream
            stop = feature.start + upstream
        else:
            start = feature.stop - upstream
            stop = feature.stop + downstream
    if add_to_name:
        try:
            feature.name += add_to_name
        except AttributeError:
            pass
    if genome is not None:
        gstart, gstop = genome[feature.chrom]
        stop = min(stop, gstop)
        start = max(start, gstart)
    if start < 0:
        start = 0
    if start > stop:
        start = stop
    feature.start = start
    feature.stop = stop
    return feature

cpdef five_prime(Interval feature, int upstream=500, int downstream=500,
                 add_to_name=None, genome=None):
    """
    Returns the 5'-most coordinate, plus `upstream` and `downstream` bp; adds
    the string `add_to_name` to the feature's name if provided (e.g., "_TSS")

    Parameters
    ----------
    feature : pybedtools.Interval instance

    upstream, downstream : int
        Number of bp upstream or downstream of the strand-specific start
        position of the feature to include. Default is 500 for both upstream
        and downstream so that the returned feature is 1kb centered on the 5'
        end of the feature. Unstranded features (where strand=".") are treated
        as plus-strand features.

    add_to_name : str or None
        If not None, append the string suffix to the name field of the feature (for
        example "_TSS").

    genome : dict or None
        If not None, then ensure that the start/stop positions are within the
        boundaries of the chromosome.
    """
    return star_prime(feature, upstream, downstream, prime=5,
                      add_to_name=add_to_name, genome=genome)


cpdef three_prime(Interval feature, int upstream=500, int downstream=500,
                  add_to_name=None, genome=None):
    """
    Returns the 3'-most coordinate, plus `upstream` and `downstream` bp; adds
    the string `add_to_name` to the feature's name if provided (e.g.,
    "_polyA_site")

    Parameters
    ----------
    feature : pybedtools.Interval instance

    upstream, downstrea : int
        Number of bp upstream or downstream of the strand-specific stop
        position of the feature to include. Default is 500 for both upstream
        and downstream so that the returned feature is 1kb centered on the 5'
        end of the feature. Unstranded features (where strand=".") are treated
        as plus-strand features.

    add_to_name : str or None
        If not None, append the string suffix to the name field of the feature (for
        example "_TSS").

    genome : dict or None
        If not None, then ensure that the start/stop positions are within the
        boundaries of the chromosome.


    """
    return star_prime(feature, upstream, downstream, prime=3,
                      add_to_name=add_to_name, genome=genome)

cpdef add_color(Interval feature, cmap, norm):
    """
    Signature:

        add_color(feature, cmap, norm)

    Given the matplotlib colormap `cmap` and the matplotlib Normalize instance
    `norm`, return a new 9-field feature (extended out if needed) with the RGB
    tuple set according to the score.
    """
    if len(feature.fields) < 9:
        feature = extend_fields(feature, 9)
        feature[6] = str(feature.start)
        feature[7] = str(feature.stop)

    rgb_float = cmap(norm(float(feature.score)))
    feature[8] = ','.join([str(int(i * 255)) for i in rgb_float[:3]])
    return feature


cpdef gff2bed(Interval feature, name_field=None):
    """
    Signature:

        gff2bed(feature, name_field=None)

    Converts a GFF feature into a BED6 feature.  By default, the name of the
    new BED will be feature.name, but if `name_field` is provided then the name
    of the new BED will be feature.attrs[name_field].

    `name_field` can also be an integer to index into the fields of the object,
    so if you want the BED name to be the GFF featuretype, then you can use
    `name_field=2`.

    If the specified field does not exist, then "." will be used for the name.
    """
    if name_field is None:
        name = feature.name
    else:
        try:
            if isinstance(name_field, basestring):
                name = feature.attrs[name_field]
            if isinstance(name_field, int):
                name = feature[name_field]
        except (NameError, KeyError):
            name = "."
    return create_interval_from_list([
        str(feature.chrom),
        str(feature.start),
        str(feature.stop),
        name,
        feature.score,
        feature.strand])


cpdef bed2gff(Interval feature):
    """
    Signature:

        bed2gff(feature)

    Converts a BED feature (BED3 through BED12) into a GFF format.

    Chrom, start, stop, score, and strand are put directly into the
    corresponding GFF fields.  Other BED fields are put into the GFF attributes
    field, named according to the UCSC BED format definition.

    If there are more than 12 BED fields, the additional fields will be added
    to the GFF attributes using the 0-based index (so starting at "12") as the
    key.

    GFF fields that do not have a direct mapping to BED format (feature type,
    source, phase) are set to ".".

    1 bp is added to the start position to finish the conversion to GFF.
    """

    # Note that Interval.score, .strand, and .name have a default of ".", so no
    # need to do the extra try/except IndexError for those fields.
    mapping = (
        (6, "thickStart"),
        (7, "thickEnd"),
        (8, "itemRgb"),
        (9, "blockCount"),
        (10, "blockSizes"),
        (11, "blockStarts")
    )

    # Add any standard BED fields we might have
    attributes = ['Name="%s"' % feature.name]
    for k, v in mapping:
        try:
            attributes.append('%s="%s"' % (v, feature.fields[k]))
        except IndexError:
            break

    # Add any additional fields, keyed by their index
    if len(feature.fields) > 12:
        for i in range(12, len(feature.fields)):
            attributes.append('%s="%s"' % (i, feature.fields[i]))

    attributes = '; '.join(attributes) + ';'

    return create_interval_from_list([
        str(feature.chrom),
        '.',
        '.',
        str(feature.start + 1),
        str(feature.stop),
        feature.score,
        feature.strand,
        '.',
        attributes])


class UniqueID(object):
    def __init__(self, pattern="%d", first=0):
        """
        Class to help create uniquely-named features.

        Example usage:

        >>> a = pybedtools.example_bedtool('a.bed')
        >>> uid = UniqueID("f_%d")
        >>> print(a.each(uid))  # doctest: +NORMALIZE_WHITESPACE
        chr1    1    100    f_0    0    +
        chr1  100    200    f_1    0    +
        chr1  150    500    f_2    0    -
        chr1  900    950    f_3    0    +

        Parameters
        ----------
        pattern : str

            Pattern will be filled in using `% self.count`

        first : int
            `self.count` will be initialzed to this value.

        """
        self.pattern = pattern
        self.first = first
        self.count = first

    def __call__(self, feature):
        feature.name = self.pattern % self.count
        self.count += 1
        return feature




# cython: profile=True

import os
from collections import deque

cdef class Window(object):
    cdef public object iterable
    cdef public object low_reads
    cdef public object high_reads
    cdef public int windowsize
    cdef int center
    cdef int left_edge
    cdef int right_edge
    cdef str chrom
    cdef public int debug
    cdef public object buffered_read
    cdef int START

    def __init__(self, iterable, windowsize=100, debug=0):
        """
        Constructor:

            Window(iterable, windowsize=100, debug=0)

        Moving window over an *iterable* of features (e.g., IntervalFile(fn)) of
        size *windowsize*.  Use *debug=1* to see all sorts of output for
        double-checking.

        The resulting Window instance can be iterated over.  Each iteration
        returns a tuple of::

            (center, low_reads, high_reads)

        where *center* is the current center of the window; *low_reads* is a
        deque of reads that includes the center and everything lower than it
        that fits within the window; and *high_reads* is a deque of reads that
        includes everything higher within the window.

        The strategy is to hold one read as the "centered read", which is
        currently in focus.  Reads are checked to see if they fit within the
        window centered on this read.  There is always a buffered read, which
        is last read taken from the iterable.  If the buffered read doesn't fit
        in the window, it remains the buffered read until the current window is
        returned.  It will continue to remain the buffered read (and no more
        reads will be taken from the iterable) until it fits within the current
        window.

        The window is implemented in two parts, a low_reads and a high_reads
        part.

        The next window's center jumps to the next available read position,
        rather than the next available bp.  This can greatly save on running
        time.  The next available read position will typically be the first
        item in the high_reads deque.

        """
        self.iterable = iterable
        self.windowsize = windowsize
        self.left_edge = 0
        self.right_edge = 0
        self.debug = debug

        # Here we pull the first thing from the iterable to set up the various
        # attributes
        first_read = self.iterable.next()
        self.chrom = first_read.chrom
        first_start_pos = first_read.start
        self.left_edge = first_start_pos - self.windowsize/2
        self.right_edge = self.left_edge + self.windowsize
        self.center = first_start_pos
        self.buffered_read = first_read
        self.high_reads = deque()
        self.low_reads = deque([self.buffered_read])
        self.START = 1

    cdef int accumulate_reads(self) except -1:
        """
        Fill up the window surrounding the currently-centered read.
        """
        if self.debug:
            print 'appending:\n\t',

        while True:


            # Need to short-circuit if starting, cause we've already filled
            # buffered_read
            if self.START:
                self.START = 0
                self.buffered_read = self.iterable.next()
                continue

            if self.buffered_read.chrom != self.chrom:
                if self.debug:
                    print 'new chrom -- %s' % self.buffered_read.chrom
                break

            # While accumulating, the only time low_reads will fill up is if
            # they are duplicates of the currently-centered read
            if self.buffered_read.start == self.center:

                if self.debug:
                    print self.buffered_read.start,

                self.low_reads.append(self.buffered_read)

            # Otherwise, if it's within the window then it's added to
            # high_reads.
            elif self.buffered_read.start < self.right_edge:

                if self.debug:
                    print  self.buffered_read.start,

                self.high_reads.append(self.buffered_read)

            else:
                break

            # The positioning of this is important -- we only get a new
            # buffered read if the last buffered read has been treated --
            # either added to low_reads or high_reads
            self.buffered_read = self.iterable.next()

        if self.debug:
            print

        return 0

    cdef int trim(self):
        """
        Trims reads off window edges, which is basically just shifting the
        window.
        """

        # If there is nothing in the high reads, then use the current buffered
        # read as the center.
        if len(self.high_reads) == 0:
            self.center = self.buffered_read.start
            self.chrom = self.buffered_read.chrom
            self.left_edge = self.center - self.windowsize/2
            self.right_edge = self.center + self.windowsize/2

        # Otherwise, use the next read in the high_reads deque
        else:
            self.chrom = self.high_reads[0].chrom
            self.center = self.high_reads[0].start
            self.left_edge = self.center - self.windowsize/2
            self.right_edge = self.center + self.windowsize/2

        # Now that the center point has been reset, remove reads from low_reads
        # list that no longer fit in the window
        if self.debug:
            print 'removed:', 
        while True:

            # Must be a better way to do this other than popping it off and
            # then back on again if it's in range, though the appendleft will
            # only happen during one (i.e. the last) time through the loop
            try:
                popped = self.low_reads.popleft()
                if (popped.start < self.left_edge) or (popped.chrom != self.buffered_read.chrom):
                    if self.debug:
                        print popped.start,
                    continue
                else:
                    self.low_reads.appendleft(popped)
                    break

            # If there's nothing left in the low_reads, then stop removing
            except IndexError:
                break

        # Next we remove any additional reads that are duplicates of the
        # centered read and add these to low_reads
        while True: 
            try:
                popped = self.high_reads.popleft()
                if popped.start == self.center:
                    self.low_reads.append(popped)
                else:
                    self.high_reads.appendleft(popped)
                    break
            except IndexError:
                break

        # Run accumulator again to see if we can add the current buffered read
        # and/or any additional reads to the window.
        #self.accumulate_reads()
        if self.debug:
            print 

        return 0

    def __iter__(self):
        return self

    def __next__(self):

        if not self.START:
            # This moves the window...
            self.trim()

        # First we accumulate reads
        self.accumulate_reads()

        if self.debug:
            print 'chrom        :', self.chrom
            print 'left         :', self.left_edge
            print 'center       :', self.center
            print 'right        :', self.right_edge
            print 'low contents :', [i.start for i in self.low_reads]
            print 'high contents:', [i.start for i in self.high_reads]
            print 'buffer       :', self.buffered_read.start

        return self.center, self.low_reads, self.high_reads





import pybedtools
import gzip
import os, difflib, sys
from textwrap import dedent
from pybedtools.helpers import BEDToolsError
from pybedtools import featurefuncs
import six
import pysam
from six.moves import socketserver
from six.moves import BaseHTTPServer
import pytest

import threading
import warnings

testdir = os.path.dirname(__file__)
tempdir = os.path.join(os.path.abspath(testdir), "tmp")
unwriteable = "unwriteable"


def setup_module():
    if not os.path.exists(tempdir):
        os.system("mkdir -p %s" % tempdir)
    pybedtools.set_tempdir(tempdir)


def teardown_module():
    if os.path.exists(tempdir):
        os.system("rm -r %s" % tempdir)
    pybedtools.cleanup()


def fix(x):
    """
    Replaces spaces with tabs, removes spurious newlines, and lstrip()s each
    line. Makes it really easy to create BED files on the fly for testing and
    checking.
    """
    s = ""
    for i in x.splitlines():
        i = i.lstrip()
        if i.endswith("\t"):
            add_tab = "\t"
        else:
            add_tab = ""
        if len(i) == 0:
            continue
        i = i.split()
        i = "\t".join(i) + add_tab + "\n"
        s += i
    return s


# ----------------------------------------------------------------------------
# Tabix support tests
# ----------------------------------------------------------------------------


def make_unwriteable():
    """
    Make a directory that cannot be written to and set the pybedtools tempdir
    to it. This is used to isolate "streaming" tests to ensure they do not
    write to disk.
    """
    if os.path.exists(unwriteable):
        os.system("rm -rf %s" % unwriteable)
    os.system("mkdir -p %s" % unwriteable)
    os.system("chmod -w %s" % unwriteable)
    pybedtools.set_tempdir(unwriteable)


def cleanup_unwriteable():
    """
    Reset to normal tempdir operation....
    """
    if os.path.exists(unwriteable):
        os.system("rm -rf %s" % unwriteable)
    pybedtools.set_tempdir(tempdir)


def test_interval_index():
    """
    supplement to the more general test in test_cbedtools.IntervalTest.testGetItemNegative
    """
    iv = pybedtools.create_interval_from_list(
        "chr21   9719768 9721892 ALR/Alpha       1004    +".split()
    )
    assert iv[-1] == "+"
    assert iv[2:-1] == ["9721892", "ALR/Alpha", "1004"]

    iv = pybedtools.create_interval_from_list(
        [
            "chr1",
            "ucb",
            "gene",
            "465",
            "805",
            ".",
            "+",
            ".",
            "ID=thaliana_1_465_805;match=scaffold_801404.1;rname=thaliana_1_465_805",
        ]
    )
    print(iv[4:-3])
    assert iv[4:-3] == ["805", "."]


def test_tuple_creation():
    # everything as a string
    t = [
        ("chr1", "1", "100", "feature1", "0", "+"),
        ("chr1", "100", "200", "feature2", "0", "+"),
        ("chr1", "150", "500", "feature3", "0", "-"),
        ("chr1", "900", "950", "feature4", "0", "+"),
    ]
    x = pybedtools.BedTool(t).saveas()
    assert pybedtools.example_bedtool("a.bed") == x

    t = [
        ("chr1", 1, 100, "feature1", 0, "+"),
        ("chr1", 100, 200, "feature2", 0, "+"),
        ("chr1", 150, 500, "feature3", 0, "-"),
        ("chr1", 900, 950, "feature4", 0, "+"),
    ]
    x = pybedtools.BedTool(t).saveas()
    assert pybedtools.example_bedtool("a.bed") == x

    t = [
        ("chr1", "fake", "gene", "50", "300", ".", "+", ".", "ID=gene1"),
        ("chr1", "fake", "mRNA", "50", "300", ".", "+", ".", "ID=mRNA1;Parent=gene1;"),
        ("chr1", "fake", "CDS", "75", "150", ".", "+", ".", "ID=CDS1;Parent=mRNA1;"),
        ("chr1", "fake", "CDS", "200", "275", ".", "+", ".", "ID=CDS2;Parent=mRNA1;"),
        ("chr1", "fake", "rRNA", "1200", "1275", ".", "+", ".", "ID=rRNA1;"),
    ]
    x = pybedtools.BedTool(t).saveas()

    # Make sure that x has actual Intervals and not plain tuples or something
    assert isinstance(x[0], pybedtools.Interval)
    assert repr(x[0]) == "Interval(chr1:49-300)"
    assert x[0]["ID"] == "gene1"


def test_tabix():
    try:
        a = pybedtools.example_bedtool("a.bed")
        t = a.tabix(force=True)
        assert t._tabixed()
        results = t.tabix_intervals("chr1:99-200")
        results = str(results)
        print(results)
        assert results == fix(
            """
        chr1	1	100	feature1	0	+
        chr1	100	200	feature2	0	+
        chr1	150	500	feature3	0	-"""
        )

        assert str(t.tabix_intervals(a[2])) == fix(
            """
        chr1	100	200	feature2	0	+
        chr1	150	500	feature3	0	-"""
        )

    finally:
        # clean up
        fns = [
            pybedtools.example_filename("a.bed.gz"),
            pybedtools.example_filename("a.bed.gz.tbi"),
        ]
        for fn in fns:
            if os.path.exists(fn):
                os.unlink(fn)


def test_tabix_intervals():
    a = pybedtools.BedTool("chr1 25 30", from_string=True).tabix()
    assert len(a.tabix_intervals("chr1:30-35")) == 0
    assert len(a.tabix_intervals("chr1:29-30")) == 1

    # make sure it works OK even if strand was provided
    assert len(a.tabix_intervals("chr1:30-35[-]")) == 0
    assert len(a.tabix_intervals("chr1:29-30[-]")) == 1

    # permit fetching of a contig without a specified region
    assert len(a.tabix_intervals("chr1")) == 1


# ----------------------------------------------------------------------------
# Streaming and non-file BedTool tests
# ----------------------------------------------------------------------------


def test_stream():
    """
    Stream and file-based equality, both whole-file and Interval by
    Interval
    """
    cleanup_unwriteable()

    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a.intersect(b)

    # this should really not be written anywhere
    d = a.intersect(b, stream=True)

    with pytest.raises(NotImplementedError):
        c.__eq__(d)
    d = d.saveas()
    d_contents = open(d.fn).read()
    c_contents = open(c.fn).read()
    assert d_contents == c_contents

    # reconstruct d and check Interval-by-Interval equality
    make_unwriteable()
    d = a.intersect(b, stream=True)

    for i, j in zip(c, d):
        assert str(i) == str(j)

    # Now do something similar with GFF files.
    a = pybedtools.example_bedtool("a.bed")
    f = pybedtools.example_bedtool("d.gff")

    # file-based
    cleanup_unwriteable()
    g1 = f.intersect(a)

    # streaming
    make_unwriteable()
    g2 = f.intersect(a, stream=True)

    for i, j in zip(g1, g2):
        assert str(i) == str(j)

    # this was segfaulting at one point, just run to make sure
    g3 = f.intersect(a, stream=True)
    for i in iter(g3):
        print(i)

    for row in a.cut([0, 1, 2, 5], stream=True):
        row[0], row[1], row[2]
        with pytest.raises(IndexError):
            row.__getitem__(4)
    cleanup_unwriteable()


def test_stream_of_stream():
    """
    Second-level streaming using self-intersections
    """
    a = pybedtools.example_bedtool("a.bed")

    # Ensure non-stream and stream equality of self-intersection
    nonstream1 = a.intersect(a, u=True)
    stream1 = a.intersect(a, u=True, stream=True)
    nonstream1_str = str(nonstream1)
    stream1_str = str(stream1)
    a_str = str(a)
    assert nonstream1_str == stream1_str == a_str

    # Have to reconstruct stream1 cause it was consumed in the str() call
    nonstream1 = a.intersect(a, u=True)
    stream1 = a.intersect(a, u=True, stream=True)
    nonstream2 = a.intersect(nonstream1, u=True)
    stream2 = a.intersect(stream1, u=True, stream=True)
    nonstream2_str = str(nonstream2)
    stream2_str = str(stream2)
    assert nonstream2_str == stream2_str == nonstream1_str == stream1_str == a_str


def test_generator():
    """
    Equality of BedTools created from file, iter(), and generator
    """
    # Test creation from file vs
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.BedTool(iter(a))
    assert str(a) == str(b)

    # Ensure that streams work well too
    b1 = a.intersect(a, stream=True)
    b2 = pybedtools.BedTool((i for i in a)).intersect(a)
    assert str(b1) == str(b2)


def test_stream_of_generator():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    b1 = a.intersect(a, stream=True)
    b2 = pybedtools.BedTool((i for i in a)).intersect(a, stream=True)
    sb1 = str(b1)
    sb2 = str(b2)
    print(sb1)
    print(sb2)
    assert sb1 == sb2


def test_many_files():
    """regression test to make sure many files can be created
    """
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    # Previously, IntervalFile would leak open files and would cause OSError
    # (too many open files) at iteration 1010 or so.
    for i in range(1100):
        c = a.intersect(b)


def test_malformed():
    """
    Malformed BED lines should raise MalformedBedLineError
    """
    a = pybedtools.BedTool(
        """
    chr1 100 200
    chr1 100 90
    chr1 100 200
    chr1 100 200
    chr1 100 200
    chr1 100 200
    """,
        from_string=True,
    )
    a_i = iter(a)

    # first feature is OK
    print(six.advance_iterator(a_i))

    # but next one is not and should raise exception
    with pytest.raises(pybedtools.MalformedBedLineError):
        a_i.__next__()


def test_remove_invalid():
    """
    Remove_invalid() removes invalid lines, track lines, and comments
    """
    a = pybedtools.BedTool(
        """
    chr1 100 200
    chr1 100 90
    track name='try to break parser'
    chr1 100 200
    chr1 100 200
    chr1 100 200
    #
    chr1 100 200
    """,
        from_string=True,
    )

    b = a.remove_invalid()

    cleaned = pybedtools.BedTool(
        """
    chr1 100 200
    chr1 100 200
    chr1 100 200
    chr1 100 200
    chr1 100 200""",
        from_string=True,
    )

    with pytest.raises(NotImplementedError):
        b.__eq__(cleaned)
    assert str(b) == str(cleaned)


def test_create_from_list_long_features():
    """
    Iterator handles extra fields from long features (BED+GFF -wao intersection)
    """
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("c.gff")
    c = a.intersect(b, wao=True, stream=False)
    d = a.intersect(b, wao=True, stream=True)

    # as of BEDTools v2.22.1, closest assumes sorted input by default.
    print(b.sort().closest(a))

    for i in d:
        print(i)


def test_iterator():
    """
    Iterator should ignore non-BED lines
    """
    s = """
    track name="test"


    browser position chrX:1-100
    # comment line
    chrX  1 10
    # more comments
    track name="another"


    """
    a = pybedtools.BedTool(s, from_string=True)
    results = list(a)
    print(results[0])
    assert str(results[0]) == "chrX\t1\t10\n", results


def test_indexing():
    """
    Indexing into BedTools
    """
    a = pybedtools.example_bedtool("a.bed")

    # This is the first line
    interval = pybedtools.Interval("chr1", 1, 100, "feature1", "0", "+")

    # just to make sure
    assert interval == next(iter(a))

    # test slice behavior
    results = list(a[0:2])
    assert len(results) == 2
    assert results[0] == interval

    # test single-integer indexing
    assert a[0] == interval

    # only slices and integers allowed....
    with pytest.raises(ValueError):
        a.__getitem__("key")


def test_repr_and_printing():
    """
    Missing files and streams should say so in repr()
    """
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a + b
    d = a.intersect(b, stream=True)
    os.unlink(c.fn)
    assert "a.bed" in repr(a)
    assert "b.bed" in repr(b)
    assert "MISSING FILE" in repr(c)
    assert "stream" in repr(d)


def test_file_type():
    """
    Regression test on file_type checks

    Previously file_type was creating a new IntervalFile every time it was
    called; now it's cached so an IntervalFile is only created once per
    BedTool.
    """
    a = pybedtools.example_bedtool("a.bed")
    for i in range(5000):
        a.file_type


# ----------------------------------------------------------------------------
# BEDTools wrapper tests --
#   See test_iter.py, which uses YAML test case definitions, for more complete
#   tests of BEDTools wrapper methods.
#
#   Here, we assert exception raises and more complicated things that can't be
#   easily described in YAML
# ----------------------------------------------------------------------------


def test_introns():
    a = pybedtools.example_bedtool("mm9.bed12")
    b = pybedtools.BedTool((f for f in a if f.name == "Tcea1,uc007afj.1")).saveas()
    bfeat = next(iter(b))

    bi = b.introns()
    # b[9] is the exonCount from teh bed12 file. there should be
    # b[9] -1 introns assuming no utrs.
    assert len(bi) == int(bfeat[9]) - 1, (len(bi), len(b))


def test_slop():
    """
    Calling slop with no genome should raise ValueError
    """
    a = pybedtools.example_bedtool("a.bed")

    # Make sure it complains if no genome is set
    with pytest.raises(ValueError):
        a.slop(**dict(l=100, r=1))


def test_closest():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    r = a.closest(b)
    assert len(r) == len(a)


# TODO: there's enough stuff in here that it's probably worth it to eventually
# make a TestSequenceStuff class
def test_sequence():
    """
    From UCSC:

    chromStart - The starting position of the feature in the chromosome or
    scaffold. The first base in a chromosome is numbered 0.

    chromEnd - The ending position of the feature in the chromosome or
    scaffold. The chromEnd base is not included in the display of the feature.
    For example, the first 100 bases of a chromosome are defined as
    chromStart=0, chromEnd=100, and span the bases numbered 0-99. """

    fi = os.path.join(testdir, "test.fasta")

    s = """
    chrX 9  16 . . +
    chrX 9  16 . . -
    chrY 1  4  . . +
    chrZ 28 31 . . +
    """

    fasta = """
    >chrX
    AAAAAAAAATGCACTGAAAAAAAAAAAAAAA
    >chrY
    GCTACCCCCCCCCCCCCCCCCCCCCCCCCCC
    >chrZ
    AAAAAAAAAAAAAAAAAAAAAAAAAAAATCT
    """
    a = pybedtools.BedTool(s, from_string=True)
    with pytest.raises(ValueError):
        a.save_seqs(("none",))

    fout = open(fi, "w")
    for line in fasta.splitlines(True):
        fout.write(line.lstrip())
    fout.close()

    # redirect stderr for the call to .sequence(), which reports the creation
    # of an index file
    tmp = open(a._tmp(), "w")
    orig_stderr = sys.stderr
    sys.stderr = tmp

    f = a.sequence(fi=fi)

    sys.stderr = orig_stderr

    assert f.fn == f.fn
    seqs = open(f.seqfn).read()
    print(seqs)
    expected = """>chrX:9-16
TGCACTG
>chrX:9-16
TGCACTG
>chrY:1-4
CTA
>chrZ:28-31
TCT
"""
    print("".join(difflib.ndiff(seqs, expected)))
    print(expected)
    assert seqs == expected

    f = a.sequence(fi=fi, s=True)
    seqs = open(f.seqfn).read()
    expected = """>chrX:9-16(+)
TGCACTG
>chrX:9-16(-)
CAGTGCA
>chrY:1-4(+)
CTA
>chrZ:28-31(+)
TCT
"""
    print(seqs)
    print(expected)
    print("".join(difflib.ndiff(seqs, expected)))
    assert seqs == expected

    f = f.save_seqs("deleteme.fa")
    assert open("deleteme.fa").read() == expected
    assert f.print_sequence() == expected
    os.unlink("deleteme.fa")

    fresh_a = pybedtools.BedTool(s, from_string=True)
    assert fresh_a == f

    os.unlink(fi)
    if os.path.exists(fi + ".fai"):
        os.unlink(fi + ".fai")


# ----------------------------------------------------------------------------
# Operator tests
# ----------------------------------------------------------------------------
def test_add_subtract():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    assert a.intersect(b, u=True) == (a + b)
    assert a.intersect(b, v=True) == (a - b)


def test_subset():
    a = pybedtools.example_bedtool("a.bed")
    import random

    random.seed(1)

    s = list(a.random_subset(1).features())
    assert len(s) == 1
    assert isinstance(s[0], pybedtools.Interval)

    s2 = list(a.random_subset(len(a)).features())
    print(len(s2))
    assert len(s2) == len(a)


def test_eq():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("a.bed")

    # BedTool to BedTool
    assert a == b

    # BedTool to string
    s = """chr1	1	100	feature1	0	+
chr1	100	200	feature2	0	+
chr1	150	500	feature3	0	-
chr1	900	950	feature4	0	+
"""
    assert a == s
    # Test not equa on bedtool
    b = pybedtools.example_bedtool("b.bed")
    assert b != a

    # and string
    assert a != "blah"

    # Don't allow testing equality on streams
    c = a.intersect(b, stream=True)
    d = a.intersect(b)
    with pytest.raises(NotImplementedError):
        c == d
    with pytest.raises(NotImplementedError):
        d == c

    # Test it on iterator, too....
    e = pybedtools.BedTool((i for i in a))
    with pytest.raises(NotImplementedError):
        a == e
    with pytest.raises(NotImplementedError):
        e == a

    # Make sure that if we force the iterator to be consumed, it is in fact
    # equal
    s = str(e)
    print(str(a).splitlines(True))
    print(s.splitlines(True))
    assert a == s


def test_hash():
    a = pybedtools.example_bedtool("a.bed")
    d = {}
    for i in a:
        d[i] = 1


# ----------------------------------------------------------------------------
# Other BedTool method tests
# ----------------------------------------------------------------------------


def test_count_bed():
    a = pybedtools.example_bedtool("a.bed")
    assert a.count() == 4
    assert len(a) == 4


def test_feature_centers():
    from pybedtools import featurefuncs

    a = pybedtools.BedTool(
        """
                           chr1 1 100
                           chr5 3000 4000
                           """,
        from_string=True,
    )
    b = a.each(featurefuncs.center, 1)
    results = list(b.features())

    print(results)

    assert results[0].start == 50
    assert results[0].stop == 51
    assert results[0].chrom == "chr1"

    assert results[1].start == 3500
    assert results[1].stop == 3501
    assert results[1].chrom == "chr5"


def test_bedtool_creation():
    # make sure we can make a bedtool from a bedtool and that it points to the
    # same file
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.BedTool(a)
    assert b.fn == a.fn
    e = FileNotFoundError if six.PY3 else ValueError
    with pytest.raises(e):
        pybedtools.BedTool("nonexistend.bed")

    # note that *s* has both tabs and spaces....
    s = """
    chr1	1	100	feature1  0	+
    chr1	100	200	feature2  0	+
    chr1	150	500	feature3  0	-
    chr1	900	950	feature4  0	+
    """
    from_string = pybedtools.BedTool(s, from_string=True)

    # difflib used here to show a bug where a newline was included when using
    # from_string
    print("".join(difflib.ndiff(str(from_string), str(a))))

    assert str(from_string) == str(a)


def test_special_methods():
    # note that *s* has both tabs and spaces....
    s = """
    chr1	1	100	feature1  0	+
    chr1	100	200	feature2  0	+
    chr1	150	500	feature3  0	-
    chr1	900	950	feature4  0	+
    """
    from_string = pybedtools.BedTool(s, from_string=True)
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")

    assert from_string == a
    assert from_string != b
    assert not from_string == b
    assert not from_string != a


def test_field_count():
    a = pybedtools.example_bedtool("a.bed")
    assert a.field_count() == 6

    tmp = pybedtools.BedTool._tmp()
    open(tmp, "w").close()
    b = pybedtools.BedTool(tmp)
    assert b.field_count() == 0


def test_repr_and_printing():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a + b
    os.unlink(c.fn)
    assert "a.bed" in repr(a)
    assert "b.bed" in repr(b)
    assert "MISSING FILE" in repr(c)

    print(a.head(1))


def test_cut():
    a = pybedtools.example_bedtool("a.bed")
    c = a.cut([0, 1, 2, 4])
    assert c.field_count() == 4, c


def test_filter():
    a = pybedtools.example_bedtool("a.bed")

    b = a.filter(lambda f: f.length < 100 and f.length > 0)
    assert len(b) == 2


def test_random_intersection():
    # TODO:
    return
    N = 4
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    li = list(a.randomintersection(b, N))
    assert len(li) == N, li


def test_cat():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    b_fn = pybedtools.example_filename("b.bed")
    assert a.cat(b) == a.cat(b_fn)
    expected = fix(
        """
    chr1 1   500
    chr1 800 950
    """
    )
    assert a.cat(b) == expected

    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a.cat(b, postmerge=False)
    assert len(a) + len(b) == len(c), (len(a), len(b), len(c))

    print(c)
    assert c == fix(
        """
    chr1	1	100	feature1	0	+
    chr1	100	200	feature2	0	+
    chr1	150	500	feature3	0	-
    chr1	900	950	feature4	0	+
    chr1	155	200	feature5	0	-
    chr1	800	901	feature6	0	+
    """
    )

    b_expected = fix(
        """
    chr1	155	200	feature5	0	-
    chr1	800	901	feature6	0	+
    """
    )
    b_merge_expected = fix(
        """
    chr1	155	200
    chr1	800	901
    """
    )
    empty = pybedtools.BedTool([])
    assert b.cat(empty) == b_merge_expected
    assert empty.cat(b) == b_merge_expected
    assert b.cat(empty, postmerge=False) == b_expected
    assert empty.cat(b, postmerge=False) == b_expected


def test_randomstats():
    chromsizes = {"chr1": (1, 1000)}
    a = pybedtools.example_bedtool("a.bed").set_chromsizes(chromsizes)
    b = pybedtools.example_bedtool("b.bed")
    try:
        results = a.randomstats(b, 100, debug=True)
        assert results["actual"] == 3
        assert results["median randomized"] == 2.0

        # NOTE: changed to reflect bedtools v2.30 output
        assert results["percentile"] == 90.0

    except ImportError:
        # allow doctests to pass if SciPy not installed
        sys.stderr.write(
            "SciPy not installed, so not testing " "BedTool.randomstats()."
        )


# ----------------------------------------------------------------------------
# Interval tests
# ----------------------------------------------------------------------------


def test_gff_stuff():
    s = """
    chr1  fake  gene 1 100 . + . ID=gene1
    chr1  fake  mRNA 1 100 . + . Name=mRNA1
    chr1  fake  CDS 50 90 . + . other=nothing
    """
    d = pybedtools.BedTool(s, from_string=True)
    f1, f2, f3 = d.features()
    assert f1.name == "gene1", f1.name
    assert f2.name == "mRNA1", f2.name
    assert f3.name is None, f3.name


def test_name():
    c = next(iter(pybedtools.example_bedtool("c.gff")))
    assert c.name == "thaliana_1_465_805", c.name


# ----------------------------------------------------------------------------
# Other non-BedTool tests
# ----------------------------------------------------------------------------


def test_flatten():
    from pybedtools.helpers import _flatten_list

    result = _flatten_list([[1, 2, 3, 0, [0, 5], 9], [100]])
    print(result)
    assert result == [1, 2, 3, 0, 0, 5, 9, 100]


def test_history_step():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a.intersect(b)
    d = c.subtract(a)

    tag = c.history[0].result_tag
    assert pybedtools.find_tagged(tag) == c

    with pytest.raises(ValueError):
        pybedtools.find_tagged("nonexistent")

    print(d.history)
    d.delete_temporary_history(ask=True, raw_input_func=lambda x: "n")
    assert os.path.exists(a.fn)
    assert os.path.exists(b.fn)
    assert os.path.exists(c.fn)
    assert os.path.exists(d.fn)

    d.delete_temporary_history(ask=True, raw_input_func=lambda x: "Yes")
    assert os.path.exists(a.fn)
    assert os.path.exists(b.fn)
    assert not os.path.exists(c.fn)  # this is the only thing that should change
    assert os.path.exists(d.fn)

    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a.intersect(b)
    d = c.subtract(a)
    d.delete_temporary_history(ask=False)
    assert os.path.exists(a.fn)
    assert os.path.exists(b.fn)
    assert not os.path.exists(c.fn)  # this is the only thing that should change
    assert os.path.exists(d.fn)


def test_kwargs():
    a = pybedtools.example_bedtool("a.bed")
    b = a.intersect(a, s=False)
    c = a.intersect(a)
    assert str(b) == str(c)


# ----------------------------------------------------------------------------
# gzip support tests
# ----------------------------------------------------------------------------


def test_is_gzip():
    gzfn = pybedtools.example_filename("snps.bed.gz")
    fn = pybedtools.example_filename("a.bed")
    assert pybedtools.helpers.isGZIP(gzfn)
    assert not pybedtools.helpers.isGZIP(fn)


def test_gzip():
    # make new gzipped files on the fly
    agz = pybedtools.BedTool._tmp()
    bgz = pybedtools.BedTool._tmp()
    os.system("gzip -c %s > %s" % (pybedtools.example_filename("a.bed"), agz))
    os.system("gzip -c %s > %s" % (pybedtools.example_filename("b.bed"), bgz))
    agz = pybedtools.BedTool(agz)
    bgz = pybedtools.BedTool(bgz)
    assert agz.file_type == bgz.file_type == "bed"
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    assert a.intersect(b) == agz.intersect(bgz) == a.intersect(bgz) == agz.intersect(b)


# ----------------------------------------------------------------------------
# BAM support tests
# ----------------------------------------------------------------------------
def test_bam_bedtool_creation():
    x = pybedtools.example_bedtool("x.bam")
    y = pybedtools.example_bedtool("y.bam")

    a = pybedtools.example_bedtool("a.bed")
    assert x._isbam
    assert y._isbam
    assert not a._isbam


def test_print_abam():
    x = pybedtools.example_bedtool("gdc.bam")
    expected = fix(
        """
    None	0	chr2L	11	255	5M	*	0	0	CGACA	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	71	255	5M	*	0	0	TTCTC	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	141	255	5M	*	0	0	CACCA	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	151	255	5M	*	0	0	GTTCA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	211	255	5M	*	0	0	AAATA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	71	255	5M	*	0	0	GAGAA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	141	255	5M	*	0	0	TGGTG	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	161	255	5M	*	0	0	GATAA	IIIII	NM:i:0	NH:i:1"""
    )
    print("x:")
    print(x)
    print("expected:")
    print(expected)
    assert x == expected


def test_bam_iter():
    x = pybedtools.example_bedtool("gdc.bam")
    s = "None	0	chr2L	11	255	5M	*	0	0	CGACA	IIIII	NM:i:0	NH:i:1\n"
    assert str(x[0]) == str(next(iter(x))) == s


# TODO: py3 branch fails here
def bam_stream_bed():
    x = pybedtools.example_bedtool("gdc.bam")
    b = pybedtools.example_bedtool("gdc.gff")
    c = x.intersect(b, u=True, bed=True, stream=True)
    str_c = str(c)
    expected = fix(
        """
    chr2L	70	75	None	255	-	70	75	0,0,0	1	5,	0,
    chr2L	140	145	None	255	-	140	145	0,0,0	1	5,	0,
    chr2L	150	155	None	255	-	150	155	0,0,0	1	5,	0,
    chr2L	210	215	None	255	+	210	215	0,0,0	1	5,	0,
    chr2L	70	75	None	255	+	70	75	0,0,0	1	5,	0,
    chr2L	140	145	None	255	+	140	145	0,0,0	1	5,	0,
    chr2L	160	165	None	255	+	160	165	0,0,0	1	5,	0,
    """
    )
    assert str_c == expected


# TODO: py3 branch fails here
def bam_stream_bam():
    x = pybedtools.example_bedtool("gdc.bam")
    b = pybedtools.example_bedtool("gdc.gff")
    c = x.intersect(b, u=True, stream=True)
    expected = fix(
        """
    None	16	chr2L	71	255	5M	*	0	0	TTCTC	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	141	255	5M	*	0	0	CACCA	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	151	255	5M	*	0	0	GTTCA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	211	255	5M	*	0	0	AAATA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	71	255	5M	*	0	0	GAGAA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	141	255	5M	*	0	0	TGGTG	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	161	255	5M	*	0	0	GATAA	IIIII	NM:i:0	NH:i:1"""
    )
    assert str(c) == expected


# TODO: py3 branch fails here
def bam_stream_bam_stream():
    x = pybedtools.example_bedtool("gdc.bam")
    b = pybedtools.example_bedtool("gdc.gff")
    c = x.intersect(b, u=True, stream=True)
    expected = fix(
        """
    None	16	chr2L	71	255	5M	*	0	0	TTCTC	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	141	255	5M	*	0	0	CACCA	IIIII	NM:i:0	NH:i:1
    None	16	chr2L	151	255	5M	*	0	0	GTTCA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	211	255	5M	*	0	0	AAATA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	71	255	5M	*	0	0	GAGAA	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	141	255	5M	*	0	0	TGGTG	IIIII	NM:i:0	NH:i:1
    None	0	chr2L	161	255	5M	*	0	0	GATAA	IIIII	NM:i:0	NH:i:1"""
    )
    d = c.intersect(b)
    print(d)
    assert str(d) == expected


def test_bam_interval():
    x = pybedtools.example_bedtool("x.bam")
    assert x[0].chrom == "chr2L"
    assert x[0].start == 9329
    assert x[0][3] == "9330"
    assert x[0].stop == 9365
    assert len(x[0][9]) == len(x[0]) == 36


def test_bam_regression():
    # Regression test:  with extra fields, the first item in x.bam was being
    # parsed as gff (cause not ==13 fields).  This does a check to prevent that
    # from happening again.
    x = pybedtools.example_bedtool("x.bam")
    assert x[0].file_type == "sam"
    assert x[0].chrom == "chr2L"


def test_sam_filetype():
    # file_type was segfaulting cause IntervalFile couldn't parse SAM
    a = pybedtools.example_bedtool("gdc.bam")
    b = pybedtools.BedTool(i for i in a).saveas()
    assert b.file_type == "sam"


def test_bam_to_sam_to_bam2():
    "test directly from #135"
    a = pybedtools.example_bedtool("gdc.bam")
    orig = str(a)
    assert a.file_type == "bam"

    # saveas should maintain BAM format
    b = a.saveas()
    assert b.file_type == "bam"

    # Converting to string gets SAM format
    assert str(b) == orig

    # b is a bam; to_bam should return a bam
    c = b.to_bam(genome="dm3")
    assert c.file_type == "bam"

    # in fact, it should be the same file:
    assert c.fn == b.fn

    # In order to get SAM format, need to print to file.
    d = open(pybedtools.BedTool._tmp(), "w")
    d.write(str(c))
    d.close()
    d = pybedtools.BedTool(d.name)
    assert d.file_type == "sam"

    e = d.to_bam(genome="dm3")
    assert e.file_type == "bam"

    # everybody should be the same
    assert a == b
    assert a == c
    assert a == d
    assert a == e


def test_bam_to_sam_to_bam():
    a = pybedtools.example_bedtool("gdc.bam")
    orig = str(a)
    assert a.file_type == "bam"

    # saveas should maintain BAM format
    b = a.saveas()
    assert b.file_type == "bam"

    # Converting to string gets SAM format
    assert str(b) == orig

    # b is a bam; to_bam should return a bam
    c = b.to_bam(genome="dm3")
    assert c.file_type == "bam"

    # in fact, it should be the same file:
    assert c.fn == b.fn

    # In order to get SAM format, need to print to file.
    d = open(pybedtools.BedTool._tmp(), "w")
    d.write(str(c))
    d.close()
    d = pybedtools.BedTool(d.name)
    assert d.file_type == "sam"

    e = d.to_bam(genome="dm3")
    assert e.file_type == "bam"

    # everybody should be the same
    assert a == b == c == d == e


def test_bam_filetype():
    # regression test -- this was segfaulting before because IntervalFile
    # couldn't parse SAM
    a = pybedtools.example_bedtool("gdc.bam")
    b = pybedtools.example_bedtool("gdc.gff")
    c = a.intersect(b)
    assert c.file_type == "bam"


def test_bam_header():
    a = pybedtools.example_bedtool("gdc.bam")
    b = pybedtools.example_bedtool("gdc.gff")
    c = a.intersect(b)
    print(c._bam_header)
    assert c._bam_header == "@SQ	SN:chr2L	LN:1800\n"


def test_output_kwarg():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a.intersect(b)
    d = a.intersect(b, output="deleteme.bed")
    assert c == d
    os.unlink("deleteme.bed")


def test_copy():
    a = pybedtools.example_bedtool("a.bed")
    x = a[0]

    # Before adding the __copy__ method to Interval class, making a copy would
    # hang and then segfault
    import copy

    y = copy.copy(x)

    assert y.start == x.start
    assert y.stop == x.stop
    assert y.chrom == x.chrom
    assert y.name == x.name
    assert y.fields == x.fields
    assert y.file_type == x.file_type == "bed"

    # Make sure it's a real copy (changing something in y doesn't change
    # something in x)
    y.start += 1
    assert y.start == x.start + 1


def test_pickleable():
    interval = pybedtools.create_interval_from_list(["chr1", "1", "100", "asdf"])
    fn = pybedtools.BedTool._tmp()
    import pickle

    out = open(fn, "wb")
    print([type(i) for i in interval.fields])
    print(type(str(interval)))
    pickle.dump(interval, out)
    out.close()
    new_interval = pickle.load(open(fn, "rb"))
    assert str(interval) == str(new_interval)

    interval = pybedtools.create_interval_from_list(["chr1", "1", "100"])
    fn = pybedtools.BedTool._tmp()
    import pickle

    out = open(fn, "wb")
    pickle.dump(interval, out)
    out.close()
    new_interval = pickle.load(open(fn, "rb"))
    assert str(interval) == str(new_interval)

    interval = pybedtools.create_interval_from_list(
        "chr2L	.	UTR	41	70	0	+	.	ID=mRNA:xs2:UTR:41-70;Parent=mRNA:xs2;".split("\t")
    )
    fn = pybedtools.BedTool._tmp()
    import pickle

    out = open(fn, "wb")
    pickle.dump(interval, out)
    out.close()
    new_interval = pickle.load(open(fn, "rb"))
    assert str(interval) == str(new_interval)


def test_split():
    a = pybedtools.example_bedtool("a.bed")

    def func(x, dist1, dist2):
        "shift the features around"

        newstart = x.start + dist1
        newstop = x.stop + dist1
        x.start = newstart
        x.stop = newstop
        yield x

        x.start -= dist2
        x.stop -= dist2

        yield x

    result = str(a.split(func, 1000, 100))
    assert result == fix(
        """
    chr1	1001	1100	feature1	0	+
    chr1	901	1000	feature1	0	+
    chr1	1100	1200	feature2	0	+
    chr1	1000	1100	feature2	0	+
    chr1	1150	1500	feature3	0	-
    chr1	1050	1400	feature3	0	-
    chr1	1900	1950	feature4	0	+
    chr1	1800	1850	feature4	0	+
    """
    )


def test_additional_args():
    a = pybedtools.example_bedtool("a.bed")
    expected = fix(
        """
    chr1	1	2	1
    chr1	100	101	1
    chr1	900	901	1"""
    )
    assert (
        a.genome_coverage(
            bg=True, strand="+", g=dict(chr1=(1, 1000)), additional_args="-5"
        )
        == expected
    )


def test_tss():
    a = pybedtools.example_bedtool("a.bed")
    results = str(
        a.each(featurefuncs.TSS, upstream=3, downstream=5, add_to_name="_TSS")
    )
    print(results)
    assert results == fix(
        """
    chr1	0	6	feature1_TSS	0	+
    chr1	97	105	feature2_TSS	0	+
    chr1	495	503	feature3_TSS	0	-
    chr1	897	905	feature4_TSS	0	+
    """
    )


def test_extend_fields():
    a = pybedtools.example_bedtool("a.bed")
    results = str(a.each(featurefuncs.extend_fields, 8))
    print(results)
    assert results == fix(
        """
    chr1	1	100	feature1	0	+	1	100
    chr1	100	200	feature2	0	+	100	200
    chr1	150	500	feature3	0	-	150	500
    chr1	900	950	feature4	0	+	900	950
    """
    )


def test_gff2bed():
    a = pybedtools.example_bedtool("d.gff")
    results = str(a.each(featurefuncs.gff2bed, name_field="Parent"))
    assert results == fix(
        """
    chr1	49	300	.	.	+
    chr1	49	300	gene1	.	+
    chr1	74	150	mRNA1	.	+
    chr1	199	275	mRNA1	.	+
    chr1	1199	1275	.	.	+"""
    )

    results = str(a.each(featurefuncs.gff2bed))
    assert results == fix(
        """
    chr1	49	300	gene1	.	+
    chr1	49	300	mRNA1	.	+
    chr1	74	150	CDS1	.	+
    chr1	199	275	CDS2	.	+
    chr1	1199	1275	rRNA1	.	+
    """
    )

    results = str(a.each(featurefuncs.gff2bed, name_field="nonexistent"))
    assert results == fix(
        """
    chr1	49	300	.	.	+
    chr1	49	300	.	.	+
    chr1	74	150	.	.	+
    chr1	199	275	.	.	+
    chr1	1199	1275	.	.	+
    """
    )

    results = str(a.each(featurefuncs.gff2bed, name_field=1))
    print(results)
    assert results == fix(
        """
    chr1	49	300	fake	.	+
    chr1	49	300	fake	.	+
    chr1	74	150	fake	.	+
    chr1	199	275	fake	.	+
    chr1	1199	1275	fake	.	+"""
    )


def test_add_color():
    try:
        from matplotlib import cm
    except ImportError:
        print("matplotlib not installed; skipping test_add_color")
        return

    def modify_scores(f):
        fields = f.fields
        fields[4] = str(f[2])
        return pybedtools.create_interval_from_list(fields)

    a = pybedtools.example_bedtool("a.bed")
    a = a.each(modify_scores).saveas()
    cmap = cm.jet
    norm = a.colormap_normalize()
    results = str(a.each(featurefuncs.add_color, cmap=cmap, norm=norm))
    print(results)
    assert results == fix(
        """
    chr1	1	100	feature1	100	+	1	100	0,0,127
    chr1	100	200	feature2	200	+	100	200	0,0,255
    chr1	150	500	feature3	500	-	150	500	99,255,147
    chr1	900	950	feature4	950	+	900	950	127,0,0"""
    )


# ------------------------------------------------------------------------------
# Tests for IntervalFile, as accessed by BedTool objects
# ------------------------------------------------------------------------------
def test_any_hits():
    a = pybedtools.example_bedtool("a.bed")

    assert 1 == a.any_hits(
        pybedtools.create_interval_from_list(["chr1", "900", "905", ".", ".", "-"])
    )

    assert 0 == a.any_hits(
        pybedtools.create_interval_from_list(["chr1", "900", "905", ".", ".", "-"]),
        same_strand=True,
    )

    assert 0 == a.any_hits(
        pybedtools.create_interval_from_list(["chr1", "8000", "9000", ".", ".", "-"])
    )


def test_all_hits():
    a = pybedtools.example_bedtool("a.bed")

    assert [a[2], a[3]] == a.all_hits(
        pybedtools.create_interval_from_list(["chr1", "450", "905", ".", ".", "-"])
    )

    assert [a[2]] == a.all_hits(
        pybedtools.create_interval_from_list(["chr1", "450", "905", ".", ".", "-"]),
        same_strand=True,
    )


def test_count_hits():
    a = pybedtools.example_bedtool("a.bed")

    assert (
        len(
            a.all_hits(
                pybedtools.create_interval_from_list(
                    ["chr1", "450", "905", ".", ".", "-"]
                )
            )
        )
        == 2
    )

    assert (
        len(
            a.all_hits(
                pybedtools.create_interval_from_list(
                    ["chr1", "450", "905", ".", ".", "-"]
                ),
                same_strand=True,
            )
        )
        == 1
    )


def test_multi_intersect():
    # Need to test here because "-i" is not a single other-bedtool like other
    # "-i" BEDTools programs, and this throws off the iter testing.
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    x = pybedtools.BedTool()
    assert x.multi_intersect(i=[a.fn, b.fn]) == fix(
        """
        chr1	1	155	1	1	1	0
        chr1	155	200	2	1,2	1	1
        chr1	200	500	1	1	1	0
        chr1	800	900	1	2	0	1
        chr1	900	901	2	1,2	1	1
        chr1	901	950	1	1	1	0"""
    )

    assert x.multi_intersect(i=[a.fn, b.fn], cluster=True) == fix(
        """
        chr1	155	200	2	1,2	1	1
        chr1	900	901	2	1,2	1	1"""
    )


def test_union_bedgraphs():
    # from unionBedGraphs -examples...

    a = pybedtools.BedTool(
        """
    chr1  1000    1500    10
    chr1  2000    2100    20
    """,
        from_string=True,
    )
    b = pybedtools.BedTool(
        """
    chr1  900 1600    60
    chr1  1700    2050    50
    """,
        from_string=True,
    )
    c = pybedtools.BedTool(
        """
    chr1  1980    2070    80
    chr1  2090    2100    20
    """,
        from_string=True,
    )

    x = pybedtools.BedTool()
    result = x.union_bedgraphs(i=[a.fn, b.fn, c.fn])
    assert result == fix(
        """
    chr1  900 1000    0   60  0
    chr1  1000    1500    10  60  0
    chr1  1500    1600    0   60  0
    chr1  1700    1980    0   50  0
    chr1  1980    2000    0   50  80
    chr1  2000    2050    20  50  80
    chr1  2050    2070    20  0   80
    chr1  2070    2090    20  0   0
    chr1  2090    2100    20  0   20
    """
    )


def test_window_maker():
    x = pybedtools.BedTool()
    a = pybedtools.example_bedtool("a.bed")
    result = x.window_maker(b=a.fn, w=50)
    print(result)
    assert result == fix(
        """
    chr1	1	51
    chr1	51	100
    chr1	100	150
    chr1	150	200
    chr1	150	200
    chr1	200	250
    chr1	250	300
    chr1	300	350
    chr1	350	400
    chr1	400	450
    chr1	450	500
    chr1	900	950
    """
    )
    x = pybedtools.BedTool()
    z = x.window_maker(genome="hg19", w=100000)
    assert str(z[0]) == "chr1\t0\t100000\n"
    assert str(z[10000]) == "chr16\t20800000\t20900000\n"


def test_random():
    a = pybedtools.BedTool()
    result = a.random(l=10, n=10, genome="hg19", seed=1).saveas()
    print(result)
    assert result == fix(
    """
    chr12	95739557	95739567	1	10	-
    chr2	113014250	113014260	2	10	-
    chr19	28057962	28057972	3	10	+
    chr4	76502010	76502020	4	10	+
    chr3	151395380	151395390	5	10	-
    chr2	43767824	43767834	6	10	+
    chr10	117350440	117350450	7	10	+
    chr6	65439870	65439880	8	10	+
    chr16	19569197	19569207	9	10	-
    chr7	104134021	104134031	10	10	-
    """
    )


def test_links():
    # have to be careful about the path, since it is embedded in the HTML
    # output -- so make a copy of the example file, and delete when done.
    os.system("cp %s a.links.bed" % pybedtools.example_filename("a.bed"))
    a = pybedtools.BedTool("a.links.bed")
    a = a.links()
    exp = open(pybedtools.example_filename("a.links.html")).read()
    obs = open(a.links_html).read()
    print(exp)
    print(obs)
    assert exp == obs
    os.unlink("a.links.bed")


def test_igv():
    a = pybedtools.example_bedtool("a.bed")
    a = a.igv()
    obs = open(a.igv_script).read()
    exp = open(pybedtools.example_filename("a.igv_script")).read()
    assert obs == exp


@pytest.mark.xfail(reason="bedtools bam2fastq (v2.30) outputs double the reads")
def test_bam_to_fastq():
    x = pybedtools.example_bedtool("small.bam")
    tmpfn = pybedtools.BedTool._tmp()
    y = x.bam_to_fastq(fq=tmpfn)
    obs = open(y.fastq).read()
    exp = open(pybedtools.example_filename("small.fastq")).read()
    print(obs)
    print(exp)

    assert (
        open(y.fastq).read() == open(pybedtools.example_filename("small.fastq")).read()
    )


def test_gtf_gff_attrs():
    # smoke test.
    #
    # this has always worked:
    gff = ["chr1", "fake", "mRNA", "51", "300", ".", "+", ".", "ID=mRNA1;Parent=gene1;"]
    gff = pybedtools.create_interval_from_list(gff)
    gff.attrs

    # this previously failed because of the "=" in the attr string.
    gff = [
        "scaffold_52",
        "Cufflinks",
        "exon",
        "5478",
        "5568",
        ".",
        "+",
        ".",
        'gene_id "XLOC_017766"; transcript_id "TCONS_00033979"; exon_number "6"; gene_name "g18412"; oId "PAC:26897502"; nearest_ref "PAC:26897502"; class_code "="; tss_id "TSS21210"; p_id "P18851";',
    ]
    gff = pybedtools.create_interval_from_list(gff)
    gff.attrs

    # TODO: is it necessary to support GFF vs GTF detection in this case:
    #
    #   GFF:
    #           class_code=" "
    #
    #   GTF:
    #           class_code "="


def test_jaccard():
    x = pybedtools.example_bedtool("a.bed")

    results = x.jaccard(pybedtools.example_bedtool("b.bed"))
    assert results == {
        "intersection": 46,
        "union": 649,
        "jaccard": 0.0708783,
        "n_intersections": 2,
    }, results

    results2 = x.jaccard(pybedtools.example_bedtool("b.bed"), stream=True)
    assert results == results2, results2


@pytest.mark.xfail
def test_reldist():
    x = pybedtools.example_bedtool("a.bed")
    results = x.reldist(pybedtools.example_bedtool("b.bed"))
    assert results == {
        "reldist": [0.15, 0.21, 0.28],
        "count": [1, 1, 1],
        "total": [3, 3, 3],
        "fraction": [0.333, 0.333, 0.333],
    }, results

    results2 = x.reldist(pybedtools.example_bedtool("b.bed"), detail=True)
    print(results2)
    assert results2 == fix(
        """
    chr1	1	100	feature1	0	+	0.282
    chr1	100	200	feature2	0	+	0.153
    chr1	150	500	feature3	0	-	0.220"""
    )


def test_remote_bam_raises_exception_when_file_doesnt_exist():
    "from #134"
    with pytest.raises(ValueError):
        pybedtools.BedTool(
            "ftp://ftp-trace.ncbi.nih.gov/this/url/clearly/does/not/exist.bam",
            remote=True,
        )


@pytest.mark.xfail(reason="Known failure: no support in BEDTools for remote BAM")
def test_remote_bam():
    url = "http://genome.ucsc.edu/goldenPath/help/examples/bamExample.bam"
    x = pybedtools.BedTool(url, remote=True)
    # for i in x:
    #    print(i)
    #    raise ValueError
    def gen():
        for i, f in enumerate(x.bam_to_bed()):
            yield f
            if i == 9:
                break

    results = pybedtools.BedTool(gen()).saveas()
    assert results == fix(
        """
    11	60636	60736	SRR081241.13799221/1	0	+
    11	60674	60774	SRR077487.5548889/1	0	+
    11	60684	60784	SRR077487.12853301/1	0	+
    11	60789	60889	SRR077487.5548889/2	0	-
    11	60950	61050	SRR077487.13826494/1	0	+
    11	60959	61059	SRR081241.13799221/2	0	-
    11	61052	61152	SRR077487.12853301/2	0	-
    11	61548	61648	SRR081241.16743804/2	0	+
    11	61665	61765	SRR081241.16743804/1	0	-
    11	61989	62089	SRR077487.167173/2	0	+"""
    ), results

    return
    # Borrow ideas from gffutils remote testing.
    #
    # Also, see issue with remote BAM hanging in pysam:
    # https://github.com/pysam-developers/pysam/issues/107
    print("Testing remote BAM by serving locally")

    class ThreadingHTTPServer(socketserver.ThreadingMixIn, BaseHTTPServer.HTTPServer):
        pass

    # handler = SimpleHTTPServer.SimpleHTTPRequestHandler
    httpd = socketserver.ThreadingTCPServer(("", 0), ThreadingHTTPServer)
    port = str(httpd.socket.getsockname()[1])
    print("Serving at port", port)

    served_folder = pybedtools.example_filename("")
    os.chdir(served_folder)
    print(served_folder)

    print("Starting SimpleHTTPServer in thread")
    server_thread = threading.Thread(target=httpd.serve_forever)
    server_thread.daemon = True
    server_thread.start()

    try:
        url = "".join(["http://localhost:", port, "/x.bam"])
        print(url)
        x = pybedtools.BedTool(url, remote=True)
        if 0:

            def gen():
                for i, f in enumerate(x.bam_to_bed(stream=True)):
                    yield f
                    if i == 9:
                        break

            results = pybedtools.BedTool(gen()).saveas()
            assert results == fix(
                """
            11	60636	60736	SRR081241.13799221/1	0	+
            11	60674	60774	SRR077487.5548889/1	0	+
            11	60684	60784	SRR077487.12853301/1	0	+
            11	60789	60889	SRR077487.5548889/2	0	-
            11	60950	61050	SRR077487.13826494/1	0	+
            11	60959	61059	SRR081241.13799221/2	0	-
            11	61052	61152	SRR077487.12853301/2	0	-
            11	61548	61648	SRR081241.16743804/2	0	+
            11	61665	61765	SRR081241.16743804/1	0	-
            11	61989	62089	SRR077487.167173/2	0	+"""
            ), results

    finally:

        print("Server shutdown.")
        httpd.shutdown()
        server_thread.join()

        # raise ValueError


def test_empty_overloaded_ops():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.BedTool("", from_string=True)
    assert b.file_type == "empty"

    # NOTE: change in semantics.  Previously, intersecting a BED file with an
    # empty file would return the original BED file.
    assert (a + b) == b
    assert (b + a) == b
    assert (a - b) == a
    assert (b - a) == b
    assert (b - b) == b


def test_to_dataframe():
    def fix_dataframe(df):
        return "".join(df.splitlines(True)[1:])

    try:
        import pandas
    except ImportError:
        pytest.xfail("pandas not installed; skipping test")

    a = pybedtools.example_bedtool("a.bed")

    results = a.to_dataframe()
    assert results.loc[0, "name"] == "feature1"
    assert list(results.columns) == ["chrom", "start", "end", "name", "score", "strand"]
    assert results.loc[3, "strand"] == "+"

    # reverse should work, too:
    df = a.to_dataframe()
    a2 = pybedtools.BedTool.from_dataframe(df)
    assert a2 == a

    # try converting only part of the dataframe to a BedTool
    a3 = pybedtools.BedTool.from_dataframe(
        df.loc[df.start < 100, ["chrom", "start", "end", "name"]]
    )
    assert a3 == fix(
        """
        chr1	1	100	feature1
        """
    ), str(a3)

    d = pybedtools.example_bedtool("d.gff")
    results = d.to_dataframe()
    assert list(results.columns) == [
        "seqname",
        "source",
        "feature",
        "start",
        "end",
        "score",
        "strand",
        "frame",
        "attributes",
    ]
    assert results.loc[0, "seqname"] == "chr1"
    assert results.loc[4, "attributes"] == "ID=rRNA1;"

    # get a gff file with too many fields...
    x = pybedtools.example_bedtool("c.gff")
    x = x.intersect(x, c=True)
    with warnings.catch_warnings(record=True) as w:
        # trigger the warning
        x.to_dataframe()
        # assert a few things
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert str(w[-1].message).startswith("Default names for filetype")

    names = [
        "seqname",
        "source",
        "feature",
        "start",
        "end",
        "score",
        "strand",
        "frame",
        "attributes",
        "count",
    ]
    results = x.to_dataframe(names=names)
    assert list(results.columns) == [
        "seqname",
        "source",
        "feature",
        "start",
        "end",
        "score",
        "strand",
        "frame",
        "attributes",
        "count",
    ]
    assert results.loc[0, "seqname"] == "chr1"
    assert results.loc[13, "count"] == 3


def test_tail():
    a = pybedtools.example_bedtool("rmsk.hg18.chr21.small.bed")
    observed = a.tail(as_string=True)
    expected = fix(
        """
        chr21	13355834	13356047	MER58A	892	-
        chr21	13356250	13356290	AT_rich	26	+
        chr21	13356358	13356381	AT_rich	23	+
        chr21	13356571	13356910	L2	333	-
        chr21	13357179	13357987	L1MEc	1264	-
        chr21	13358003	13358300	L1MEc	379	-
        chr21	13358304	13358952	L1MEc	1271	-
        chr21	13358960	13359288	L2	336	+
        chr21	13359444	13359751	AluY	2337	+
        chr21	13360044	13360225	L1M5	284	-"""
    )
    assert observed == expected

    # only ask for 3 lines
    observed = a.tail(3, as_string=True)
    expected = fix(
        """
        chr21	13358960	13359288	L2	336	+
        chr21	13359444	13359751	AluY	2337	+
        chr21	13360044	13360225	L1M5	284	-"""
    )
    assert observed == expected

    # For short files, whole thing should be returned
    a = pybedtools.example_bedtool("a.bed")
    expected = str(a)
    obs = a.tail(as_string=True)
    assert obs == expected


def test_fisher():
    a = pybedtools.example_bedtool("a.bed")
    b = pybedtools.example_bedtool("b.bed")
    c = a.fisher(b, genome="hg19")
    assert (
        str(c)
        == """# Number of query intervals: 4
# Number of db intervals: 2
# Number of overlaps: 3
# Number of possible intervals (estimated): 13958448
# phyper(3 - 1, 4, 13958448 - 4, 2, lower.tail=F)
# Contingency Table Of Counts
#_________________________________________
#           |  in -b       | not in -b    |
#     in -a | 3            | 1            |
# not in -a | 0            | 13958444     |
#_________________________________________
# p-values for fisher's exact test
left	right	two-tail	ratio
1	8.8247e-21	8.8247e-21	inf
"""
    ), c


def test_zero_len_boolean():
    # regression test: with no __nonzero__ method, __len__ is used. This means
    # that `if interval` evaluates to False when length is zero.
    i = pybedtools.create_interval_from_list(["chr1", "1", "1"])
    assert len(i) == 0
    assert i


def test_chromsizes_in_5prime_3prime():
    # standard 5'
    a = (
        pybedtools.example_bedtool("a.bed")
        .each(
            featurefuncs.five_prime,
            1,
            10,
            add_to_name="_TSS",
            genome=pybedtools.chromsizes("hg19"),
        )
        .saveas()
    )
    assert a == fix(
        """
        chr1	0	11	feature1_TSS	0	+
        chr1	99	110	feature2_TSS	0	+
        chr1	490	501	feature3_TSS	0	-
        chr1	899	910	feature4_TSS	0	+
        """
    ), str(a)

    # add genomes sizes; last feature should be truncated
    a = (
        pybedtools.example_bedtool("a.bed")
        .each(
            featurefuncs.five_prime,
            1,
            10,
            add_to_name="_TSS",
            genome=dict(chr1=(0, 900)),
        )
        .saveas()
    )
    assert a == fix(
        """
        chr1	0	11	feature1_TSS	0	+
        chr1	99	110	feature2_TSS	0	+
        chr1	490	501	feature3_TSS	0	-
        chr1	899	900	feature4_TSS	0	+
        """
    ), str(a)

    # same thing but for 3'.
    # Note that the last feature chr1:949-960 is completely truncated because
    # it would entirely fall outside of the chromosome
    a = (
        pybedtools.example_bedtool("a.bed")
        .each(
            featurefuncs.three_prime,
            1,
            10,
            add_to_name="_TSS",
            genome=dict(chr1=(0, 900)),
        )
        .saveas()
    )
    assert a == fix(
        """
        chr1	99	110	feature1_TSS	0	+
        chr1	199	210	feature2_TSS	0	+
        chr1	140	151	feature3_TSS	0	-
        chr1	900	900	feature4_TSS	0	+
        """
    ), str(a)

    # be a lot harsher with the chromsizes to ensure features on both strands
    # get truncated correctly
    a = (
        pybedtools.example_bedtool("a.bed")
        .each(
            featurefuncs.three_prime,
            1,
            10,
            add_to_name="_TSS",
            genome=dict(chr1=(0, 120)),
        )
        .saveas()
    )
    assert a == fix(
        """
        chr1	99	110	feature1_TSS	0	+
        chr1	120	120	feature2_TSS	0	+
        chr1	120	120	feature3_TSS	0	-
        chr1	120	120	feature4_TSS	0	+
        """
    ), str(a)


def test_new_head():
    """
    The new BedTool.head no longer iterates using IntervalIterator but instead
    just prints the lines of the file directly.
    """
    tmp = pybedtools.BedTool._tmp()
    with open(tmp, "w") as fout:
        fout.write("chr1\t5\t10\n" "chr1\t-1\t15\n")
    a = pybedtools.BedTool(tmp)

    # previously would crash with OverflowError, can't convert negative value
    # to CHRPOS
    a.head()

    # however, printing should still complain:
    with pytest.raises(pybedtools.cbedtools.MalformedBedLineError):
        print(a)



import sys
import multiprocessing
from . import helpers
import pybedtools


def _parallel_wrap(
    orig_bedtool,
    shuffle_kwargs,
    genome_fn,
    method,
    method_args,
    method_kwargs,
    sort=False,
    shuffle=True,
    reduce_func=None,
):
    """
    Given a BedTool object `orig_bedtool`, call its `method` with `args` and
    `kwargs` and then call `reduce_func` on the results.

    See parallel_apply docstring for details

    """

    # note: be careful about cleaning up tempfiles
    if not shuffle:
        to_use = orig_bedtool
    else:
        shuffled = orig_bedtool.shuffle(g=genome_fn, **shuffle_kwargs)
        if sort:
            to_use = shuffled.sort()
            helpers.close_or_delete(shuffled)
        else:
            to_use = shuffled

    result = getattr(to_use, method)(*method_args, **method_kwargs)

    if shuffle:
        helpers.close_or_delete(to_use)

    if reduce_func:
        reduced = reduce_func(result)
        if isinstance(result, pybedtools.BedTool):
            helpers.close_or_delete(result)
        return reduced
    else:
        return result


def parallel_apply(
    orig_bedtool,
    method,
    genome=None,
    genome_fn=None,
    method_args=None,
    method_kwargs=None,
    shuffle_kwargs=None,
    shuffle=True,
    reduce_func=None,
    processes=1,
    sort=False,
    _orig_pool=None,
    iterations=1000,
    debug=False,
    report_iterations=False,
):
    """
    Call an arbitrary BedTool method many times in parallel.

    An example use-case is to generate a null distribution of intersections,
    and then compare this to the actual intersections.

    **Important:** due to a known file handle leak in BedTool.__len__, it's
    best to simply check the number of lines in the file, as in the below
    function. This works because BEDTools programs strip any non-interval lines
    in the results.

    >>> # set up example BedTools
    >>> a = pybedtools.example_bedtool('a.bed')
    >>> b = pybedtools.example_bedtool('b.bed')

    >>> # Method of `a` to call:
    >>> method = 'intersect'

    >>> # Kwargs provided to `a.intersect` each iteration
    >>> method_kwargs = dict(b=b, u=True)

    >>> # Function that will be called on the results of
    >>> # `a.intersect(**method_kwargs)`.
    >>> def reduce_func(x):
    ...     return sum(1 for _ in open(x.fn))

    >>> # Create a small artificial genome for this test (generally you'd
    >>> # use an assembly name, like "hg19"):
    >>> genome = dict(chr1=(0, 1000))

    >>> # Do 10 iterations using 1 process for this test (generally you'd
    >>> # use 1000+ iterations, and as many processes as you have CPUs)
    >>> results = pybedtools.parallel.parallel_apply(a, method, genome=genome,
    ... method_kwargs=method_kwargs, iterations=10, processes=1,
    ... reduce_func=reduce_func, debug=True, report_iterations=True)

    >>> # get results
    >>> print(list(results))
    [1, 0, 1, 2, 4, 2, 2, 1, 2, 4]

    >>> # We can compare this to the actual intersection:
    >>> reduce_func(a.intersect(**method_kwargs))
    3

    Alternatively, we could use the `a.jaccard` method, which already does the
    reduction to a dictionary.  However, the Jaccard method requires the input
    to be sorted.  Here, we specify `sort=True` to sort each shuffled BedTool
    before calling its `jaccard` method.

    >>> from pybedtools.parallel import parallel_apply
    >>> a = pybedtools.example_bedtool('a.bed')
    >>> results = parallel_apply(a, method='jaccard', method_args=(b,),
    ... genome=genome, iterations=3, processes=1, sort=True, debug=True)
    >>> for i in results:
    ...     print(sorted(i.items()))
    [('intersection', 12), ('jaccard', 0.0171184), ('n_intersections', 1), ('union', 701)]
    [('intersection', 0), ('jaccard', 0.0), ('n_intersections', 0), ('union', 527)]
    [('intersection', 73), ('jaccard', 0.137996), ('n_intersections', 1), ('union', 529)]

    Parameters
    ----------
    orig_bedtool : BedTool

    method : str
        The method of `orig_bedtool` to run

    method_args : tuple
        Passed directly to getattr(orig_bedtool, method)()

    method_kwargs : dict
        Passed directly to getattr(orig_bedtool, method)()

    shuffle : bool
        If True, then `orig_bedtool` will be shuffled at each iteration and
        that shuffled version's `method` will be called with `method_args` and
        `method_kwargs`.

    shuffle_kwargs : dict
        If `shuffle` is True, these are passed to `orig_bedtool.shuffle()`.
        You do not need to pass the genome here; that's handled separately by
        the `genome` and `genome_fn` kwargs.

    iterations : int
        Number of iterations to perform

    genome : string or dict
        If string, then assume it is the assembly name (e.g., hg19) and get
        a dictionary of chromsizes for that assembly, then converts to
        a filename.

    genome_fn : str
        Mutually exclusive with `genome`; `genome_fn` must be an existing
        filename with the chromsizes.  Use the `genome` kwarg instead if you'd
        rather supply an assembly or dict.

    reduce_func : callable
        Function or other callable object that accepts, as its only argument,
        the results from `orig_bedtool.method()`.  For example, if you care
        about the number of results, then you can use `reduce_func=len`.

    processes : int
        Number of processes to run.  If `processes=1`, then multiprocessing is
        not used (making it much easier to debug).  This argument is ignored if
        `_orig_pool` is provided.

    sort : bool
        If both `shuffle` and `sort` are True, then the shuffled BedTool will
        then be sorted.  Use this if `method` requires sorted input.

    _orig_pool : multiprocessing.Pool instance
        If provided, uses `_orig_pool` instead of creating one.  In this case,
        `processes` will be ignored.

    debug : bool
        If True, then use the current iteration index as the seed to shuffle.

    report_iterations : bool
        If True, then report the number of iterations to stderr.
    """

    shuffle_kwargs = shuffle_kwargs or {}
    method_args = method_args or ()
    if not isinstance(method_args, list) and not isinstance(method_args, tuple):
        raise ValueError(
            "method_args must be a list or tuple, got %s" % type(method_args)
        )
    method_kwargs = method_kwargs or {}

    if genome_fn and genome:
        raise ValueError("only of of genome_fn or genome should be provided")

    if shuffle:
        if not genome_fn:
            if not genome:
                raise ValueError(
                    "shuffle=True, so either genome_fn" " or genome must be provided"
                )
            genome_fn = pybedtools.chromsizes_to_file(genome)

    _parallel_wrap_kwargs = dict(
        orig_bedtool=orig_bedtool,
        shuffle_kwargs=shuffle_kwargs,
        genome_fn=genome_fn,
        method=method,
        method_args=method_args,
        method_kwargs=method_kwargs,
        shuffle=shuffle,
        reduce_func=reduce_func,
        sort=sort,
    )

    def add_seed(i, kwargs):
        if debug and shuffle:
            kwargs["shuffle_kwargs"]["seed"] = i
        return kwargs

    if processes == 1:
        for it in range(iterations):
            yield _parallel_wrap(**add_seed(it, _parallel_wrap_kwargs))
        return

    if _orig_pool:
        p = _orig_pool
    else:
        p = multiprocessing.Pool(processes)

    results = [
        p.apply_async(_parallel_wrap, (), add_seed(it, _parallel_wrap_kwargs))
        for it in range(iterations)
    ]
    for i, r in enumerate(results):
        yield r.get()
        if report_iterations:
            sys.stderr.write("%s\r" % i)
            sys.stderr.flush()



#!/usr/bin/env python

"""
Example from pybedtools documentation: find reads in introns and exons using
multiple CPUs.

Prints a tab-separated file containing class (exon, intron, both) and number of
reads in each class.
"""
import pybedtools
import argparse
import os
import sys
import multiprocessing

if __name__ == "__main__":

    ap = argparse.ArgumentParser(prog=os.path.basename(sys.argv[0]), usage=__doc__)
    ap.add_argument(
        "--gff", required=True, help="GFF or GTF file containing annotations"
    )
    ap.add_argument(
        "--bam", required=True, help="BAM file containing reads to be counted"
    )
    ap.add_argument(
        "--stranded",
        action="store_true",
        help="Use strand-specific merging and overlap. " "Default is to ignore strand",
    )
    ap.add_argument(
        "--processes",
        default=1,
        type=int,
        help="Number of processes to use in parallel.",
    )
    ap.add_argument(
        "-v", "--verbose", action="store_true", help="Verbose (goes to stderr)"
    )
    args = ap.parse_args()

    gff = args.gff
    bam = args.bam
    stranded = args.stranded

    if args.processes > 3:
        print(
            "Only need 3 processes (one each for exon, intron, both), so "
            "resetting processes from {0} to 3".format(args.processes)
        )
        args.processes = 3

    def featuretype_filter(feature, featuretype):
        """
        Only passes features with the specified *featuretype*
        """
        if feature[2] == featuretype:
            return True
        return False

    def subset_featuretypes(featuretype):
        """
        Returns the filename containing only `featuretype` features.
        """
        return g.filter(featuretype_filter, featuretype).saveas().fn

    def count_reads_in_features(features):
        """
        Callback function to count reads in features
        """
        return (
            pybedtools.BedTool(bam).intersect(
                features, s=stranded, bed=True, stream=True
            )
        ).count()

    # Some GFF files have invalid entries -- like chromosomes with negative coords
    # or features of length = 0.  This line removes them (the `remove_invalid`
    # method) and saves the result in a tempfile
    g = pybedtools.BedTool(gff).remove_invalid().saveas()

    # Set up pool of workers
    with multiprocessing.Pool(processes=args.processes) as pool:

        # Get separate files for introns and exons in parallel
        featuretypes = ["intron", "exon"]
        introns, exons = pool.map(subset_featuretypes, featuretypes)

        # Since `subset_featuretypes` returns filenames, we convert to BedTool objects
        # to do intersections below.
        introns = pybedtools.BedTool(introns)
        exons = pybedtools.BedTool(exons)

        # Identify unique and shared regions using bedtools commands subtract, merge,
        # and intersect.
        exon_only = exons.subtract(introns).merge()
        intron_only = introns.subtract(exons).merge()
        intron_and_exon = exons.intersect(introns).merge()

        # Do intersections with BAM file in parallel. Note that we're passing filenames
        # to multiprocessing.Pool rather than BedTool objects.
        features = (exon_only.fn, intron_only.fn, intron_and_exon.fn)

        # Run count_reads_in_features in parallel over features
        results = pool.map(count_reads_in_features, features)

        labels = ("exon_only", "intron_only", "intron_and_exon")

    for label, reads in zip(labels, results):
        print("{0}\t{1}".format(label, reads))



import difflib
import copy
import itertools
import yaml
import os
import gzip
import pybedtools
from .tfuncs import setup_module, teardown_module

yamltestdesc = ["test_cases.yaml"]

if pybedtools.settings._v_2_27_plus:
    yamltestdesc.append("test_merge227.yaml")
    yamltestdesc.append("test_shuffle227.yaml")

elif pybedtools.settings._v_2_15_plus and not pybedtools.settings._v_2_27_plus:
    yamltestdesc.append("test_merge215.yaml")
    yamltestdesc.append("test_shuffle215.yaml")

this_dir = os.path.dirname(__file__)
yamltestdesc = [os.path.join(this_dir, i) for i in yamltestdesc]


def gz(x):
    """
    Gzips a file to a tempfile, and returns a new BedTool using the gzipped
    version.
    """
    gzfn = pybedtools.BedTool._tmp()
    with gzip.open(gzfn, "wb") as out_:
        with open(x.fn, "rb") as in_:
            out_.writelines(in_)
    return pybedtools.BedTool(gzfn)


def fix(x):
    """
    Replaces spaces with tabs, removes spurious newlines, and lstrip()s each
    line. Makes it really easy to create BED files on the fly for testing and
    checking.
    """
    s = ""
    for i in x.splitlines():
        i = i.strip("\n\r")
        if len(i) == 0:
            continue

        # If the expected output contains tabs, then use those to split,
        # otherwise space.  This allows you to have expected output with blank
        # fields (e.g., "\t\t")
        if "\t" in i:
            i = i.split("\t")
        else:
            i = i.split()

        i = "\t".join(i) + "\n"
        s += i
    return s


# List of methods that *only* take BAM as input
bam_methods = ("bam_to_bed",)

# List of supported BedTool construction from BAM files.  Currently only
# file-based.
supported_bam = ("filename",)

converters = {
    "filename": lambda x: pybedtools.BedTool(x.fn),
    "generator": lambda x: pybedtools.BedTool(i for i in x),
    "stream": lambda x: pybedtools.BedTool(open(x.fn)),
    "gzip": gz,
}


def run(d):
    method = d["method"]
    bedtool = d["bedtool"]
    convert = d["convert"]
    kwargs = d["kw"].copy()
    expected = d["test_case"]["expected"]

    bedtool_converter = convert.pop("bedtool")
    bedtool = converters[bedtool_converter](pybedtools.example_bedtool(bedtool))

    for k, converter_name in convert.items():
        kwargs[k] = converters[converter_name](pybedtools.example_bedtool(kwargs[k]))
    result = getattr(bedtool, method)(**kwargs)
    res = str(result)
    expected = fix(expected)
    try:
        assert res == expected

    except AssertionError:
        print(result.fn)
        print("Method call:")
        args = []
        for key, val in list(kwargs.items()):
            args.append(("%s=%s" % (key, val)).strip())

        args = ", ".join(args)
        print("BedTool.%(method)s(%(args)s)" % locals())
        print("Got:")
        print(res)
        print("Expected:")
        print(expected)
        print("Diff:")
        for i in difflib.unified_diff(res.splitlines(1), expected.splitlines(1)):
            print(i, end=" ")

        # Make tabs and newlines visible
        spec_res = res.replace("\t", "\\t").replace("\n", "\\n\n")
        spec_expected = expected.replace("\t", "\\t").replace("\n", "\\n\n")

        print("Showing special characters:")
        print("Got:")
        print(spec_res)
        print("Expected:")
        print(spec_expected)
        print("Diff:")
        for i in difflib.unified_diff(
            spec_res.splitlines(1), spec_expected.splitlines(1)
        ):
            print(i, end=" ")
        raise


def pytest_generate_tests(metafunc):
    tests = []
    labels = []
    for config_fn in yamltestdesc:
        if hasattr(yaml, "FullLoader"):
            test_cases = yaml.load(open(config_fn).read(), Loader=yaml.FullLoader)
        else:
            test_cases = yaml.load(open(config_fn).read())
        for test_case in test_cases:
            kw = test_case["kwargs"]

            kwc = copy.copy(kw)

            method = test_case["method"]

            a_isbam = False
            b_isbam = False
            i_isbam = False

            # Figure out if this is a test involving a method that operates on
            # two files (ab), one file (i) or the "bed" tests (bed)
            flavor = None
            if (("a" in kw) and ("b" in kw)) or ("abam" in kw):
                flavor = "ab"
            if ("i" in kw) or ("ibam" in kw):
                flavor = "i"

            # If bams were specified in the test block we need to keep track of
            # that. Then we set the 'a' or 'i' kws, which control the nature of
            # the constructed BedTool, to that filename.
            if "abam" in kw:
                kw["abam"] = pybedtools.example_filename(kw["abam"])
                kw["a"] = kw["abam"]
                a_isbam = True

            if "ibam" in kw:
                kw["ibam"] = pybedtools.example_filename(kw["ibam"])
                kw["i"] = kw["ibam"]
                i_isbam = True

            kinds = ["filename", "generator", "stream", "gzip"]

            if "files" in kw:
                kw["files"] = [pybedtools.example_filename(i) for i in kw["files"]]

            if "bams" in kw:
                kw["bams"] = [pybedtools.example_filename(i) for i in kw["bams"]]

            if "fi" in kw:
                kw["fi"] = pybedtools.example_filename(kw["fi"])

            if flavor == "i":
                orig_i = pybedtools.example_bedtool(kw["i"])
                if orig_i._isbam:
                    i_isbam = True
                bedtool = kw.pop("i")
                for kind in kinds:
                    if i_isbam and (kind not in supported_bam):
                        continue
                    label = "{method}: {kwc} {kind}".format(**locals())
                    labels.append(label)
                    tests.append(
                        dict(
                            method=method,
                            bedtool=bedtool,
                            test_case=test_case,
                            kw=kw,
                            convert={"bedtool": kind},
                        )
                    )

            if flavor == "ab":
                orig_a = pybedtools.example_bedtool(kw["a"])
                orig_b = pybedtools.example_bedtool(kw["b"])
                if orig_a._isbam:
                    a_isbam = True
                if orig_b._isbam:
                    b_isbam = True

                bedtool = kw.pop("a")

                for kind_a, kind_b in itertools.permutations(kinds, 2):
                    label = "{method}: {kwc} a={kind_a} b={kind_b}".format(**locals())
                    if a_isbam and (kind_a not in supported_bam):
                        continue
                    if b_isbam and (kind_b not in supported_bam):
                        continue

                    labels.append(label)
                    tests.append(
                        dict(
                            method=method,
                            bedtool=bedtool,
                            test_case=test_case,
                            kw=kw,
                            convert={"bedtool": kind_a, "b": kind_b},
                        )
                    )

    metafunc.parametrize("tests", tests, ids=labels)


def test_all(tests):
    run(tests)



_bedtools_path = ""
_R_path = ""

tempfile_prefix = "pybedtools."
tempfile_suffix = ".tmp"

# Checking for BEDTools will happen when creating the first BedTool; other
# checks happen at first use (BAM object creation; tabix-ing a BedTool)
_bedtools_installed = False
_R_installed = False
_v_2_15_plus = False
_v_2_27_plus = False
bedtools_version = []

KEEP_TEMPFILES = False
_DEBUG = True

# Check calls against these names to only allow calls to known BEDTools
# programs (basic security)
#
_prog_names = {
    # Genome arithmetic
    "intersectBed": "intersect",
    "windowBed": "window",
    "closestBed": "closest",
    "coverageBed": "coverage",
    "mapBed": "map",
    "genomeCoverageBed": "genomecov",
    "mergeBed": "merge",
    "clusterBed": "cluster",
    "complementBed": "complement",
    "subtractBed": "subtract",
    "slopBed": "slop",
    "shiftBed": "shift",
    "flankBed": "flank",
    "sortBed": "sort",
    "randomBed": "random",
    "shuffleBed": "shuffle",
    "annotateBed": "annotate",
    "spacing": "spacing",
    # multi-way
    "multiIntersectBed": "multiinter",
    "unionBedGraphs": "unionbedg",
    # PE
    "pairToBed": "pairtobed",
    "pairToPair": "pairtopair",
    # format conversion
    "bamToBed": "bamtobed",
    "bedToBam": "bedtobam",
    "bedpeToBam": "bedpetobam",
    "bed12ToBed6": "bed12tobed6",
    "bamToFastq": "bamtofastq",
    # fasta
    "fastaFromBed": "getfasta",
    "maskFastaFromBed": "maskfasta",
    "nucBed": "nuc",
    # bam-centric
    "multiBamCov": "multicov",
    "tagBam": "tag",
    # stats
    "jaccard": "jaccard",
    "reldist": "reldist",
    # misc
    "getOverlap": "overlap",
    "bedToIgv": "igv",
    "linksBed": "links",
    "windowMaker": "makewindows",
    "groupBy": "groupby",
    "expandCols": "expand",
    "sample": "sample",
    "fisher": "fisher",
    "split": "split",
}

_old_names = list(_prog_names.keys())
_new_names = list(_prog_names.values())

_column_names = {
    "bed": [
        "chrom",
        "start",
        "end",
        "name",
        "score",
        "strand",
        "thickStart",
        "thickEnd",
        "itemRgb",
        "blockCount",
        "blockSizes",
        "blockStarts",
    ],
    "gff": [
        "seqname",
        "source",
        "feature",
        "start",
        "end",
        "score",
        "strand",
        "frame",
        "attributes",
    ],
}



def suite():
    import unittest
    import os

    test_dir = os.path.abspath(os.path.dirname(__file__))

    tests = []
    for filename in os.listdir(test_dir):
        if filename.endswith(".py") and not filename.startswith("_"):
            tests.append(unittest.findTestCases("tests." + filename[:-3]))

    return unittest.TestSuite(tests)


if __name__ == "__main__":
    import unittest

    unittest.main()



#!/usr/bin/python
"""
Create a matrix of many pairwise intersections; see \
:mod:`pybedtools.contrib.IntersectionMatrix` for more flexibility
"""

import collections
import time
import sys
import os.path as op
import argparse
import pybedtools
from pybedtools import BedTool, example_filename

usage = (
    """

    Send in a list of `N` bed files, and this script will create an N by
    N matrix of their intersections, or optionally, co-localization scores.

    Run the example with::

        %s --test > matrix.txt

    You can then plot a quick heatmap in R with::

        > m = as.matrix(read.table("matrix.txt"))
        > heatmap(m)

"""
    % sys.argv[0]
)


def get_name(fname):
    return op.splitext(op.basename(fname))[0]


def actual_intersection(a, b):
    return len(a.intersect(b, u=True))


def frac_of_a(a, b):
    len_a = float(len(a))
    return len(a.intersect(b, u=True)) / len_a


def enrichment_score(a, b, genome_fn, iterations=None, processes=None):
    results = a.randomstats(
        b, new=True, genome_fn=genome_fn, iterations=iterations, processes=processes
    )
    return (results["actual"] + 1) / (results["median randomized"] + 1)


def create_matrix(beds, func, verbose=False, **kwargs):
    nfiles = len(beds)
    total = nfiles ** 2
    i = 0
    matrix = collections.defaultdict(dict)
    for fa in beds:
        a = BedTool(fa)
        for fb in beds:
            i += 1
            b = BedTool(fb)

            if verbose:
                sys.stderr.write("%(i)s of %(total)s: %(fa)s + %(fb)s\n" % locals())
                sys.stderr.flush()

            matrix[get_name(fa)][get_name(fb)] = func(a, b, **kwargs)

    return matrix


def main():
    """
    Creates a pairwise matrix containing overlapping feature counts for many
    BED files
    """
    ap = argparse.ArgumentParser(usage=usage)
    ap.add_argument(
        "beds",
        nargs="*",
        help="BED/GTF/GFF/VCF filenames, e.g., "
        "in a directory of bed files, you can use *.bed",
    )
    ap.add_argument(
        "--frac",
        action="store_true",
        help="Instead of counts, report fraction overlapped",
    )
    ap.add_argument(
        "--enrichment",
        action="store_true",
        help="Run randomizations (default 1000, specify otherwise "
        "with --iterations) on each pairwise comparison and "
        "compute the enrichment score as "
        "(actual intersection count + 1) / (median randomized + 1)",
    )
    ap.add_argument(
        "--genome",
        help="Required argument if --enrichment is "
        'used. Needs to be a string assembly name like "dm3" or '
        '"hg19"',
    )
    ap.add_argument(
        "--iterations",
        default=1000,
        type=int,
        help="Number of randomizations to perform for enrichement " "scores",
    )
    ap.add_argument(
        "--processes",
        default=None,
        type=int,
        help="Number of CPUs to use for randomization",
    )
    ap.add_argument(
        "--test",
        action="store_true",
        help="Ignore any input BED " "files and use test BED files",
    )
    ap.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Be verbose: print which files are "
        "currently being intersected and timing info at the end.",
    )
    args = ap.parse_args()

    if not args.beds and not args.test:
        ap.print_help()
        sys.exit(1)

    if args.test:
        # insulator binding sites from ChIP-chip -- 4 proteins, 2 cell types
        # Genes Dev. 2009 23(11):1338-1350
        args.beds = [
            example_filename(i)
            for i in [
                "Cp190_Kc_Bushey_2009.bed",
                "Cp190_Mbn2_Bushey_2009.bed",
                "CTCF_Kc_Bushey_2009.bed",
                "CTCF_Mbn2_Bushey_2009.bed",
                "SuHw_Kc_Bushey_2009.bed",
                "SuHw_Mbn2_Bushey_2009.bed",
                "BEAF_Mbn2_Bushey_2009.bed",
                "BEAF_Kc_Bushey_2009.bed",
            ]
        ]

    if args.enrichment:
        FUNC = enrichment_score
        genome_fn = pybedtools.chromsizes_to_file(pybedtools.chromsizes(args.genome))
        kwargs = dict(
            genome_fn=genome_fn, iterations=args.iterations, processes=args.processes
        )

    elif args.frac:
        FUNC = frac_of_a
        kwargs = {}
    else:
        FUNC = actual_intersection
        kwargs = {}

    t0 = time.time()
    matrix = create_matrix(beds=args.beds, func=FUNC, verbose=args.verbose, **kwargs)
    t1 = time.time()

    nfiles = len(args.beds)

    if args.verbose:
        sys.stderr.write(
            "Time to construct %s x %s matrix: %.1fs" % (nfiles, nfiles, (t1 - t0))
            + "\n"
        )
    keys = sorted(matrix.keys())

    sys.stdout.write("\t" + "\t".join(keys) + "\n")
    for k in keys:
        sys.stdout.write(k)
        for j in keys:
            sys.stdout.write("\t" + str(matrix[k][j]))
        sys.stdout.write("\n")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Given 3 files, creates a 3-way Venn diagram of intersections using the Google \
Chart API; see :mod:`pybedtools.contrib.venn_maker` for more flexibility.

The values in the diagram assume:

    * unstranded intersections
    * no features that are nested inside larger features
"""

import argparse
import sys
import pybedtools
from six.moves import urllib


def venn_gchart(a, b, c=None, colors=None, labels=None, size="300x300"):
    """
    a, b, and c are filenames to BED-like files.

    *colors* is a list of 3 hex colors

    *labels* is a list of 3 labels

    *outfn* is the output PNG you want to create.

    *size* is the size in pixels for the PNG
    """
    a = pybedtools.BedTool(a)
    b = pybedtools.BedTool(b)
    if c:
        c = pybedtools.BedTool(c)

    # The order of values is meaningful to the API, see
    # http://code.google.com/apis/chart/docs/gallery/venn_charts.html
    if c:
        vals = [
            len(a),
            len(b),
            len(c),
            len(a + b),
            len(a + c),
            len(b + c),
            len(a + b + c),
        ]
    else:
        # insert 0 for size of 3rd circle.
        vals = [len(a), len(b), 0, len(a + b)]
        labels = labels[:2]
    # API doesn't seem to like large numbers, so get fractions instead, then
    # join make a comma-separated list of values.
    mx = float(max(vals))
    vals = [round(i / mx, 4) for i in vals]
    valstr = ",".join(map(str, vals))

    data = {"cht": "v", "chs": size, "chd": "t:" + valstr}

    # Add the optional data, if specified
    if labels:
        data["chdl"] = "|".join(labels)
    if colors:
        data["chco"] = ",".join(colors)
    return data


def gchart(data, outfn="out.png"):
    """
    Sends data to Google Chart API
    """
    data = urllib.parse.urlencode(data)
    binary_data = data.encode("UTF-8")

    url = "https://chart.googleapis.com/chart?"

    # Request and get the PNG
    req = urllib.request.Request(url, binary_data)
    print(url + data)
    response = urllib.request.urlopen(req)
    f = open(outfn, "w")
    f.write(response.read())
    f.close()


def main():
    """Create a 3-way Venn diagram using Google Charts API
    """

    op = argparse.ArgumentParser(description=__doc__, prog=sys.argv[0])
    op.add_argument("-a", help="File to use for the left-most circle")
    op.add_argument("-b", help="File to use for the right-most circle")
    op.add_argument("-c", help="File to use for the bottom circle")
    op.add_argument(
        "--colors",
        help="Optional comma-separated list of hex "
        "colors for circles a, b, and c.  E.g. %(default)s",
        default="00FF00,FF0000,0000FF",
    )
    op.add_argument(
        "--labels",
        help="Optional comma-separated list of labels for a, b, and c",
        default="a,b,c",
    )
    op.add_argument(
        "--size",
        default="300x300",
        help="Optional size of PNG, in pixels.  Default is " '"%(default)s"',
    )
    op.add_argument(
        "-o", default="out.png", help="Output file to save as, in PNG format"
    )
    op.add_argument(
        "--test", action="store_true", help="run test, overriding all other options."
    )
    options = op.parse_args()

    reqd_args = ["a", "b"]
    if not options.test:
        for ra in reqd_args:
            if not getattr(options, ra):
                op.print_help()
                sys.stderr.write('Missing required arg "%s"\n' % ra)
                sys.exit(1)

    if options.test:
        # Example data
        pybedtools.bedtool.random.seed(1)
        a = pybedtools.example_bedtool("rmsk.hg18.chr21.small.bed")
        b = pybedtools.example_bedtool("venn.b.bed")
        c = pybedtools.example_bedtool("venn.c.bed")
        options.a = a.fn
        options.b = b.fn
        options.c = c.fn
        options.colors = "00FF00,FF0000,0000FF"
        options.o = "out.png"
        options.labels = "a,b,c"

    data = venn_gchart(
        a=options.a,
        b=options.b,
        c=options.c,
        colors=options.colors.split(","),
        labels=options.labels.split(","),
        size=options.size,
    )
    gchart(data, outfn=options.o)


if __name__ == "__main__":
    import doctest

    if doctest.testmod(optionflags=doctest.ELLIPSIS).failed == 0:
        main()



import multiprocessing
from . import helpers
import pybedtools


def random_jaccard(x, y, genome_fn, shuffle_kwargs, jaccard_kwargs):
    z = x.shuffle(g=genome_fn, **shuffle_kwargs).sort()
    result = z.jaccard(y, **jaccard_kwargs)
    helpers.close_or_delete(z)
    return result


def random_intersection(x, y, genome_fn, shuffle_kwargs, intersect_kwargs):
    z = x.shuffle(g=genome_fn, **shuffle_kwargs)
    zz = z.intersect(y, stream=True, **intersect_kwargs)
    result = len(zz)
    helpers.close_or_delete(z, zz)
    return result


def random_intersection_bp(x, y, genome_fn, shuffle_kwargs, intersect_kwargs):
    z = x.shuffle(g=genome_fn, **shuffle_kwargs)
    zz = z.intersect(y, stream=True, **intersect_kwargs)
    result = sum(len(i) for i in zz)
    helpers.close_or_delete(z, zz)
    return result



from six.moves import builtins
import os
import sys
import pytest


# Make sure all tests import pybedtools + genomepy anew
@pytest.fixture(autouse=True)
def cleanup_imports():
    yield
    del_mods = [mod for mod in sys.modules if "pybedtools" in mod or "genomepy" in mod]
    for mod in del_mods:
        del sys.modules[mod]


def test_genomepy_not_installed():
    import pybedtools

    if "genomepy" in sys.modules:
        del sys.modules["genomepy"]
    genome = "pybedtools/test/data/genome.fa"
    d = pybedtools.helpers.get_chromsizes_from_genomepy(genome)
    assert d is None

    with pytest.raises(OSError):
        pybedtools.chromsizes(genome)
    with pytest.raises(OSError):
        pybedtools.chromsizes("non-existing")


def test_chromsizes_from_genomepy():
    import pybedtools

    if "genomepy" not in sys.modules:
        pytest.skip("genomepy not instlled -- skipping test")

    genome = "pybedtools/test/data/genome.fa"
    try:
        d = pybedtools.helpers.get_chromsizes_from_genomepy(genome)
        print(pybedtools.chromsizes(genome))
        assert d["chr1"] == (0, 10)

        d = pybedtools.chromsizes(genome)
        assert d["chr3"] == (0, 30)

    finally:
        # Make sure all genomepy files get deleted
        fnames = [genome + ext for ext in [".fai", ".sizes"]]
        fnames.append(genome.replace(".fa", ".gaps.bed"))
        for fname in fnames:
            if os.path.exists(fname):
                os.unlink(fname)

    assert None == pybedtools.helpers.get_chromsizes_from_genomepy("non-existing")



import pytest
import pybedtools
import os

testdir = os.path.dirname(__file__)
test_tempdir = os.path.join(os.path.abspath(testdir), "tmp")
unwriteable = os.path.join(os.path.abspath(testdir), "unwriteable")


def setup_module():
    if not os.path.exists(test_tempdir):
        os.system("mkdir -p %s" % test_tempdir)
    pybedtools.set_tempdir(test_tempdir)


def teardown_module():
    if os.path.exists(test_tempdir):
        os.system("rm -r %s" % test_tempdir)
    pybedtools.cleanup()



"""
when bugs are identified post-release, put tests here to make sure they don't
happen again
"""

import pybedtools
import pybedtools.featurefuncs
import pybedtools.helpers


def test_midpoint():
    """
    regression test for #98
    """
    a = """chr1 3052874 3053149
    chr1 3333690 3333915
    chr1 3472838 3473382
    chr1 3639053 3639356
    """

    def nothing(f):
        return f

    input_bed = pybedtools.BedTool(a, from_string=True).saveas("test_input.bed")

    for func in [
        pybedtools.featurefuncs.midpoint,
        pybedtools.featurefuncs.center,
        nothing,
    ]:
        input_bed_mid = input_bed.each(func)
        assert len(input_bed_mid) == 4


# pysam is now handling bgzip
# def test_bgzip_missing():
#
#     old_path = pybedtools.settings._tabix_path
#     pybedtools.helpers.set_bgzip_path('somenonexistantpath')
#     a = pybedtools.example_bedtool('a.bed')
#     assert_raises(ValueError, a.tabix)
#     pybedtools.helpers.set_bgzip_path(old_path)



import os
import sys
import itertools
import six
import time
import pysam
import pybedtools


def tag_bedpe(bedpe, queries, verbose=False):
    """
    Tag each end of a BEDPE with a set of (possibly many) query BED files.

    For example, given a BEDPE of interacting fragments from a Hi-C experiment,
    identify the contacts between promoters and ChIP-seq peaks. In this case,
    promoters and ChIP-seq peaks of interest would be provided as BED files.

    The strategy is to split the BEDPE into two separate files.  Each file is
    intersected independently with the set of queries.  The results are then
    iterated through in parallel to tie the ends back together. It is this
    iterator that is returned (see example below).

    Parameters
    ----------

    bedpe : str
        BEDPE-format file. Must be name-sorted.

    queries : dict
        Dictionary of BED/GFF/GTF/VCF files to use. After splitting the BEDPE,
        these query files (values in the dictionary) will be passed as the `-b`
        arg to `bedtools intersect`. The keys are passed as the `names`
        argument for `bedtools intersect`

        *Features in each file must have unique names*. Use
        :func:`pybedtools.featurefuncs.UniqueID` to help fix this.

        *Each file must be BED3 to BED6*.

    Returns
    -------
    Tuple of (iterator, n, extra).

    `iterator` is described below. `n` is the total number of lines in the
    BEDPE file, which is useful for calculating percentage complete for
    downstream work. `extra` is the number of extra fields found in the BEDPE
    (also useful for downstream processing).

    `iterator` yields tuples of (label, end1_hits, end2_hits) where `label` is
    the name field of one line of the original BEDPE file. `end1_hits` and
    `end2_hits` are each iterators of BED-like lines representing all
    identified intersections across all query BED files for end1 and end2 for
    this pair.

    Recall that BEDPE format defines a single name and a single score for each
    pair. For each item in `end1_hits`, the fields are::

        chrom1
        start1
        end1
        name
        score
        strand1
        [extra fields]
        query_label
        fields_from_query_intersecting_end1

    where `[extra fields]` are any additional fields from the original BEDPE,
    `query_label` is one of the keys in the `beds` input dictionary, and the
    remaining fields in the line are the intersecting line from the
    corresponding BED file in the `beds` input dictionary.

    Similarly, each item in `end2_hits` consists of:

        chrom2
        start2
        end2
        name
        score
        strand2
        [extra fields]
        query_label
        fields_from_query_intersecting_end2

    At least one line is reported for every line in the BEDPE file. If there
    was no intersection, the standard BEDTools null fields will be shown. In
    `end1_hits` and `end2_hits`, a line will be reported for each hit in each
    query.

    Example
    -------

    Consider the following BEDPE (where "x1" is an aribtrary extra field).

    >>> bedpe = pybedtools.example_bedtool('test_bedpe.bed')
    >>> print(bedpe) # doctest: +NORMALIZE_WHITESPACE
    chr1  1  10  chr1  50   90   pair1  5  +  -  x1
    chr1  2  15  chr1  200  210  pair2  1  +  +  y1
    <BLANKLINE>


    And the following transcription start sites (TSSes) in BED4 format:

    >>> tsses = pybedtools.example_bedtool('test_tsses.bed')
    >>> print(tsses) # doctest: +NORMALIZE_WHITESPACE
    chr1  5   6   gene1
    chr1  60  61  gene2
    chr1  88  89  gene3
    <BLANKLINE>

    And the following called peaks as BED6:

    >>> peaks = pybedtools.example_bedtool('test_peaks.bed')
    >>> print(peaks) # doctest: +NORMALIZE_WHITESPACE
    chr1  3  4  peak1  50  .
    <BLANKLINE>

    Then we can get the following iterator, n, and extra. Note that the
    OrderedDict is only for testing to ensure output is always consistend; in
    practice a regular dictionary is fine:

    >>> from pybedtools.contrib.long_range_interaction import tag_bedpe
    >>> from collections import OrderedDict
    >>> queries = OrderedDict()
    >>> queries['tss'] = tsses
    >>> queries['pk'] = peaks
    >>> iterator, n, extra = tag_bedpe(bedpe, queries)
    >>> print(n)
    2
    >>> print(extra)
    1

    The following illustrates that each item in the iterator represents one
    pair, and each item in each group represents an intersection with one end.
    Note that the sorting is necessary only for the doctests to be output in
    consistent format; this not typically needed:

    >>> for (label, end1_hits, end2_hits) in iterator:
    ...    end1_hits = sorted(end1_hits, key=lambda x: str(x))
    ...    end2_hits = sorted(end2_hits, key=lambda x: str(x))
    ...    print('PAIR = {}'.format(label))
    ...    print('end1_hits:')
    ...    for i in end1_hits:
    ...        print(i, end='')
    ...    print('end2_hits:')
    ...    for i in end2_hits:
    ...        print(i, end='')  # doctest: +NORMALIZE_WHITESPACE
    PAIR = pair1
    end1_hits:
    chr1       1       10      pair1   5       +       x1      pk      chr1    3       4       peak1   50      .       1
    chr1       1       10      pair1   5       +       x1      tss     chr1    5       6       gene1   1
    end2_hits:
    chr1       50      90      pair1   5       -       x1      tss     chr1    60      61      gene2   1
    chr1       50      90      pair1   5       -       x1      tss     chr1    88      89      gene3   1
    PAIR = pair2
    end1_hits:
    chr1       2       15      pair2   1       +       y1      pk      chr1    3       4       peak1   50      .       1
    chr1       2       15      pair2   1       +       y1      tss     chr1    5       6       gene1   1
    end2_hits:
    chr1       200     210     pair2   1       +       y1      .       .       -1      -1      .       0

    See the `cis_trans_interactions()` function for one way of summarizing
    these data.
    """
    b = pybedtools.BedTool(bedpe)

    # Figure out if the supplied bedpe had any extra fields. If so, the fields
    # are repeated in each of the split output files.
    observed = b.field_count()
    extra = observed - 10
    extra_inds = [10 + i for i in range(extra)]

    end1_fn = pybedtools.BedTool._tmp()
    end2_fn = pybedtools.BedTool._tmp()

    # Performance notes:
    # We don't need the overhead of converting every line into
    # a pybedtools.Interval object just so we can grab the fields. Doing so
    # takes 3.5x more time than simply splitting each line on a tab.
    if verbose:
        print("splitting BEDPE into separate files.")
        print("end1 is going to %s" % end1_fn)
        print("end2 is going to %s" % end2_fn)

    n = 0
    with open(end1_fn, "w") as end1_out, open(end2_fn, "w") as end2_out:
        for line in open(b.fn):
            n += 1
            f = line.strip().split("\t")
            end1_out.write(
                "\t".join((f[i] for i in [0, 1, 2, 6, 7, 8] + extra_inds)) + "\n"
            )
            end2_out.write(
                "\t".join((f[i] for i in [3, 4, 5, 6, 7, 9] + extra_inds)) + "\n"
            )

    # Performance notes:
    #
    # For small BEDPE and large set of query files, it would be faster to sort
    # these independently, intersect with sorted=True, and then re-sort by name
    # for the grouping. For large BEDPE, I don't think the sorted=True
    # performance gain outweighs the hit from sorting twice.
    #
    # On the other hand, if BEDPE was coord-sorted in the first place, only
    # end2 would need to be sorted and re-sorted. On the other (third!?) hand,
    # BEDPE creation from BAM implies name-sorting, so it's probably not
    # reasonable to assume coord-sorted.
    #
    # In the end: don't do any sorting.

    end1_bt = pybedtools.BedTool(end1_fn)
    end2_bt = pybedtools.BedTool(end2_fn)
    names, fns = [], []
    for name, fn in queries.items():
        names.append(name)
        if isinstance(fn, pybedtools.BedTool):
            fns.append(fn.fn)
        else:
            fns.append(fn)

    if verbose:
        print("intersecting end 1")
    end1_hits = end1_bt.intersect(list(fns), names=names, wao=True)
    if verbose:
        print("intersecting end 2")
    end2_hits = end2_bt.intersect(list(fns), names=names, wao=True)
    if verbose:
        print("intersection with end1 is in %s" % (end1_hits.fn))
        print("intersection with end2 is in %s" % (end2_hits.fn))

    grouped_end1 = itertools.groupby(end1_hits, lambda f: f[3])
    grouped_end2 = itertools.groupby(end2_hits, lambda f: f[3])

    def gen():
        for (label1, group1), (label2, group2) in six.moves.zip(
            grouped_end1, grouped_end2
        ):
            assert label1 == label2
            yield label1, group1, group2

    return gen(), n, extra


def cis_trans_interactions(iterator, n, extra, verbose=True):
    """
    Converts the output from `tag_bedpe` into a pandas DataFrame containing
    information about regions that contact each other in cis (same fragment) or
    trans (different fragments).

    For example, given a BEDPE file representing 3D interactions in the genome,
    we want to identify which transcription start sites are connected to distal
    regions containing a peak.

    >>> bedpe = pybedtools.example_bedtool('test_bedpe.bed')
    >>> print(bedpe) # doctest: +NORMALIZE_WHITESPACE
    chr1  1  10  chr1  50   90   pair1  5  +  -  x1
    chr1  2  15  chr1  200  210  pair2  1  +  +  y1
    <BLANKLINE>

    >>> tsses = pybedtools.example_bedtool('test_tsses.bed')
    >>> print(tsses) # doctest: +NORMALIZE_WHITESPACE
    chr1  5   6   gene1
    chr1  60  61  gene2
    chr1  88  89  gene3
    <BLANKLINE>

    >>> peaks = pybedtools.example_bedtool('test_peaks.bed')
    >>> print(peaks) # doctest: +NORMALIZE_WHITESPACE
    chr1  3  4  peak1  50  .
    <BLANKLINE>

    Here's what the tracks look like. Note that pair1 is evidence of
    a gene1-gene2 interaction and a gene1-gene3 interaction::

        TRACKS:

                      1         2  /         5         6    / 8         9     / 20
            0123456789012345678901 / 2345678901234567890123 / 012345678901234 / 0123456789
       pair1 |||||||||------------ / --------|||||||||||||| / ||||||||||||||| /
       pair2  |||||||||||||------- / ---------------------- / --------------- / ||||||||||
       tsses     1                 /                   2    /         3
       peaks   1


    >>> from collections import OrderedDict
    >>> queries = OrderedDict()
    >>> queries['tss'] = tsses
    >>> queries['pk'] = peaks
    >>> iterator, n, extra = tag_bedpe(bedpe, queries)
    >>> for (label, group1, group2) in iterator:
    ...    group1 = sorted(group1, key=lambda x: str(x))
    ...    group2 = sorted(group2, key=lambda x: str(x))
    ...    for i in group1:
    ...        print(i, end='')  # doctest: +NORMALIZE_WHITESPACE
    ...    for i in group2:
    ...        print(i, end='')  # doctest: +NORMALIZE_WHITESPACE
    chr1       1       10      pair1   5       +       x1      pk      chr1    3       4       peak1   50      .       1
    chr1       1       10      pair1   5       +       x1      tss     chr1    5       6       gene1   1
    chr1       50      90      pair1   5       -       x1      tss     chr1    60      61      gene2   1
    chr1       50      90      pair1   5       -       x1      tss     chr1    88      89      gene3   1
    chr1       2       15      pair2   1       +       y1      pk      chr1    3       4       peak1   50      .       1
    chr1       2       15      pair2   1       +       y1      tss     chr1    5       6       gene1   1
    chr1       200     210     pair2   1       +       y1      .       .       -1      -1      .       0

    Now we run the same thing, but now aggregate it. Note that each piece of
    interaction evidence has its own line. The first line shows that pair1 has
    gene1 and peak1 in the same fragment, and that they are connected to gene2.
    The second line shows again that gene1 and peak1 are in the same fragmet
    and that they are also connected to gene3:

    >>> import pandas; pandas.set_option('display.max_columns', 10)
    >>> iterator, n, extra = tag_bedpe(bedpe, {'tss': tsses, 'pk': peaks})
    >>> df =  cis_trans_interactions(iterator, n, extra)
    >>> print(df.sort_values(list(df.columns)).reset_index(drop=True))
      target_label target_name cis_label cis_name distal_label distal_name  label
    0           pk       peak1       tss    gene1            .           .  pair2
    1           pk       peak1       tss    gene1          tss       gene2  pair1
    2           pk       peak1       tss    gene1          tss       gene3  pair1
    3          tss       gene1        pk    peak1            .           .  pair2
    4          tss       gene1        pk    peak1          tss       gene2  pair1
    5          tss       gene1        pk    peak1          tss       gene3  pair1
    6          tss       gene2       tss    gene3           pk       peak1  pair1
    7          tss       gene2       tss    gene3          tss       gene1  pair1
    8          tss       gene3       tss    gene2           pk       peak1  pair1
    9          tss       gene3       tss    gene2          tss       gene1  pair1

    If we only care about genes:

    >>> print((df[df.target_label == 'tss']).sort_values(list(df.columns)).reset_index(drop=True))
      target_label target_name cis_label cis_name distal_label distal_name  label
    0          tss       gene1        pk    peak1            .           .  pair2
    1          tss       gene1        pk    peak1          tss       gene2  pair1
    2          tss       gene1        pk    peak1          tss       gene3  pair1
    3          tss       gene2       tss    gene3           pk       peak1  pair1
    4          tss       gene2       tss    gene3          tss       gene1  pair1
    5          tss       gene3       tss    gene2           pk       peak1  pair1
    6          tss       gene3       tss    gene2          tss       gene1  pair1


    Note that in pair2, there is no evidence of interaction between gene1 and
    gene2.

    What interacts distally with gene2's TSS?

    >>> assert set(df.loc[df.target_name == 'gene2', 'distal_name']).difference('.') == set([u'gene1', u'peak1'])

    """
    try:
        import pandas
    except ImportError:
        raise ImportError("pandas must be installed to use this function")
    c = 0
    lines = []
    for label, end1_hits, end2_hits in iterator:
        c += 1
        if c % 1000 == 0:
            print("%d (%.1f%%)\r" % (c, c / float(n) * 100), end="")
            sys.stdout.flush()

        # end1_hits has the full lines of all intersections with end1
        end1_hits = list(end1_hits)
        end2_hits = list(end2_hits)

        def get_name_hits(f):
            """
            Returns the key (from which file the interval came) and the name
            (of the individual feature).
            """
            # this is the "name" reported if there was no hit.
            if f[6 + extra] == ".":
                return (".", ".")
            interval = pybedtools.create_interval_from_list(f[7 + extra :])
            return [f[6 + extra], interval.name]

        names1 = set(six.moves.map(tuple, six.moves.map(get_name_hits, end1_hits)))
        names2 = set(six.moves.map(tuple, six.moves.map(get_name_hits, end2_hits)))

        for cis, others in [(names1, names2), (names2, names1)]:
            for target in cis:
                if target == (".", "."):
                    continue
                non_targets = set(cis).difference([target])
                if len(non_targets) == 0:
                    non_targets = [(".", ".")]
                for non_target in non_targets:
                    for other in others:
                        line = []
                        line.extend(target)
                        line.extend(non_target)
                        line.extend(other)
                        line.append(label)
                        lines.append(line)

    df = pandas.DataFrame(
        lines,
        columns=[
            "target_label",
            "target_name",
            "cis_label",
            "cis_name",
            "distal_label",
            "distal_name",
            "label",
        ],
    )
    df = df.drop_duplicates()
    return df
